import json
import uuid
from typing import Optional
from unittest import IsolatedAsyncioTestCase

import pytest
from marshmallow import ValidationError
from uuid_utils import uuid4

from ...connections.models.conn_record import ConnRecord
from ...ledger.base import BaseLedger
from ...ledger.endpoint_type import EndpointType
from ...ledger.multiple_ledger.base_manager import BaseMultipleLedgerManager
from ...ledger.multiple_ledger.ledger_config_schema import (
    ConfigurableWriteLedgersSchema,
    LedgerConfigInstanceSchema,
    LedgerConfigListSchema,
)
from ...ledger.multiple_ledger.ledger_requests_executor import IndyLedgerRequestsExecutor
from ...multitenant.base import BaseMultitenantManager
from ...multitenant.manager import MultitenantManager
from ...tests import mock
from ...utils.testing import create_test_profile
from .. import routes as test_module
from ..indy_vdr import Role


class TestLedgerRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.ledger = mock.create_autospec(BaseLedger)
        self.ledger.pool_name = "pool.0"
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.context = self.profile.context
        setattr(self.context, "profile", self.profile)
        self.profile.context.injector.bind_instance(BaseLedger, self.ledger)
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

        self.test_did = "did"
        self.test_verkey = "verkey"
        self.test_endpoint = "http://localhost:8021"
        self.test_endpoint_type = EndpointType.PROFILE
        self.test_endpoint_type_profile = "http://company.com/profile"
        self.mock_ledger_requests_executor = mock.MagicMock(
            IndyLedgerRequestsExecutor, autospec=True
        )

    async def test_missing_ledger(self):
        self.mock_ledger_requests_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, None)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor,
            self.mock_ledger_requests_executor,
        )
        self.profile.context.injector.clear_binding(BaseLedger)
        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.register_ledger_nym(self.request)

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.get_nym_role(self.request)

        with self.assertRaises(test_module.web.HTTPForbidden):
            self.request.query["did"] = "test"
            await test_module.get_nym_role(self.request)

        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.rotate_public_did_keypair(self.request)

        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.get_did_verkey(self.request)

        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.get_did_endpoint(self.request)

        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.ledger_accept_taa(self.request)

        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.ledger_get_taa(self.request)

    async def test_get_verkey_a(self):
        self.mock_ledger_requests_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor,
            self.mock_ledger_requests_executor,
        )
        self.request.query = {"did": self.test_did}
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.ledger.get_key_for_did.return_value = self.test_verkey
            result = await test_module.get_did_verkey(self.request)
            json_response.assert_called_once_with(
                {"verkey": self.ledger.get_key_for_did.return_value}
            )
            assert result is json_response.return_value

    async def test_get_verkey_b(self):
        self.mock_ledger_requests_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=("test_ledger_id", self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor,
            self.mock_ledger_requests_executor,
        )
        self.request.query = {"did": self.test_did}
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.ledger.get_key_for_did.return_value = self.test_verkey
            result = await test_module.get_did_verkey(self.request)
            json_response.assert_called_once_with(
                {
                    "ledger_id": "test_ledger_id",
                    "verkey": self.ledger.get_key_for_did.return_value,
                }
            )
            assert result is json_response.return_value

    async def test_get_verkey_multitenant(self):
        self.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        self.request.query = {"did": self.test_did}
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor,
            self.mock_ledger_requests_executor,
        )
        with (
            mock.patch.object(
                IndyLedgerRequestsExecutor,
                "get_ledger_for_identifier",
                mock.CoroutineMock(return_value=("test_ledger_id", self.ledger)),
            ),
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as json_response,
        ):
            self.ledger.get_key_for_did.return_value = self.test_verkey
            result = await test_module.get_did_verkey(self.request)
            json_response.assert_called_once_with(
                {
                    "ledger_id": "test_ledger_id",
                    "verkey": self.ledger.get_key_for_did.return_value,
                }
            )
            assert result is json_response.return_value

    async def test_get_verkey_no_did(self):
        self.request.query = {"no": "did"}
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.get_did_verkey(self.request)

    async def test_get_verkey_did_not_public(self):
        self.mock_ledger_requests_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=("test_ledger_id", self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor,
            self.mock_ledger_requests_executor,
        )

        self.request.query = {"did": self.test_did}
        self.ledger.get_key_for_did.return_value = None
        with self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.get_did_verkey(self.request)

    async def test_get_verkey_x(self):
        self.mock_ledger_requests_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor,
            self.mock_ledger_requests_executor,
        )
        self.request.query = {"did": self.test_did}
        self.ledger.get_key_for_did.side_effect = test_module.LedgerError()
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.get_did_verkey(self.request)

    async def test_get_endpoint(self):
        self.mock_ledger_requests_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor,
            self.mock_ledger_requests_executor,
        )
        self.request.query = {"did": self.test_did}
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.ledger.get_endpoint_for_did.return_value = self.test_endpoint
            result = await test_module.get_did_endpoint(self.request)
            json_response.assert_called_once_with(
                {"endpoint": self.ledger.get_endpoint_for_did.return_value}
            )
            assert result is json_response.return_value

    async def test_get_endpoint_multitenant(self):
        self.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        self.request.query = {"did": self.test_did}
        with (
            mock.patch.object(
                IndyLedgerRequestsExecutor,
                "get_ledger_for_identifier",
                mock.CoroutineMock(return_value=("test_ledger_id", self.ledger)),
            ),
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as json_response,
        ):
            self.ledger.get_endpoint_for_did.return_value = self.test_endpoint
            result = await test_module.get_did_endpoint(self.request)
            json_response.assert_called_once_with(
                {
                    "ledger_id": "test_ledger_id",
                    "endpoint": self.ledger.get_endpoint_for_did.return_value,
                }
            )
            assert result is json_response.return_value

    async def test_get_endpoint_of_type_profile(self):
        self.mock_ledger_requests_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=("test_ledger_id", self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor,
            self.mock_ledger_requests_executor,
        )

        self.request.query = {
            "did": self.test_did,
            "endpoint_type": self.test_endpoint_type.w3c,
        }
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.ledger.get_endpoint_for_did.return_value = (
                self.test_endpoint_type_profile
            )
            result = await test_module.get_did_endpoint(self.request)
            json_response.assert_called_once_with(
                {
                    "ledger_id": "test_ledger_id",
                    "endpoint": self.ledger.get_endpoint_for_did.return_value,
                }
            )
            assert result is json_response.return_value

    async def test_get_endpoint_no_did(self):
        self.request.query = {"no": "did"}
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.get_did_endpoint(self.request)

    async def test_get_endpoint_x(self):
        self.mock_ledger_requests_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=("test_ledger_id", self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor,
            self.mock_ledger_requests_executor,
        )

        self.request.query = {"did": self.test_did}
        self.ledger.get_endpoint_for_did.side_effect = test_module.LedgerError()
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.get_did_endpoint(self.request)

    async def test_register_nym(self):
        self.request.query = {
            "did": self.test_did,
            "verkey": self.test_verkey,
            "role": "reset",
        }
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            success: bool = True
            txn: Optional[dict] = None
            self.ledger.register_nym.return_value = (success, txn)
            result = await test_module.register_ledger_nym(self.request)
            json_response.assert_called_once_with({"success": success})
            assert result is json_response.return_value

    async def test_register_nym_bad_request(self):
        self.request.query = {"no": "did"}
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.register_ledger_nym(self.request)

    async def test_register_nym_ledger_txn_error(self):
        self.request.query = {"did": self.test_did, "verkey": self.test_verkey}
        self.ledger.register_nym.side_effect = test_module.LedgerTransactionError("Error")
        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.register_ledger_nym(self.request)

    async def test_register_nym_ledger_error(self):
        self.request.query = {"did": self.test_did, "verkey": self.test_verkey}
        self.ledger.register_nym.side_effect = test_module.LedgerError("Error")
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.register_ledger_nym(self.request)

    async def test_register_nym_wallet_not_found_error(self):
        self.request.query = {"did": self.test_did, "verkey": self.test_verkey}
        self.ledger.register_nym.side_effect = test_module.WalletNotFoundError("Error")
        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.register_ledger_nym(self.request)

    async def test_register_nym_wallet_error(self):
        self.request.query = {"did": self.test_did, "verkey": self.test_verkey}
        self.ledger.register_nym.side_effect = test_module.WalletError("Error")
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.register_ledger_nym(self.request)

    async def test_register_nym_create_transaction_for_endorser(self):
        self.request.query = {
            "did": "a_test_did",
            "verkey": "a_test_verkey",
            "alias": "did_alias",
            "role": "ENDORSER",
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        serialize=mock.MagicMock(return_value={"...": "..."})
                    )
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "endorser_did": ("did"),
                        "endorser_name": ("name"),
                    }
                )
            )
            self.ledger.register_nym.return_value = (
                True,
                {"signed_txn": {"...": "..."}},
            )

            result = await test_module.register_ledger_nym(self.request)
            assert result == mock_response.return_value
            mock_response.assert_called_once_with(
                {"success": True, "txn": {"signed_txn": {"...": "..."}}}
            )

    async def test_register_nym_create_transaction_for_endorser_no_public_did(self):
        self.request.query = {
            "did": "a_test_did",
            "verkey": "a_test_verkey",
            "alias": "did_alias",
            "role": "reset",
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }
        self.profile.context.settings["endorser.author"] = True

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        serialize=mock.MagicMock(return_value={"...": "..."})
                    )
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "endorser_did": ("did"),
                        "endorser_name": ("name"),
                    }
                )
            )
            self.ledger.register_nym.return_value = (
                True,
                {"signed_txn": {"...": "..."}},
            )

            result = await test_module.register_ledger_nym(self.request)
            assert result == mock_response.return_value
            mock_response.assert_called_once_with(
                {"success": True, "txn": {"signed_txn": {"...": "..."}}}
            )

    async def test_register_nym_create_transaction_for_endorser_storage_x(self):
        self.request.query = {
            "did": "a_test_did",
            "verkey": "a_test_verkey",
            "alias": "did_alias",
            "role": "ENDORSER",
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_record=mock.CoroutineMock(side_effect=test_module.StorageError())
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "endorser_did": ("did"),
                        "endorser_name": ("name"),
                    }
                )
            )
            self.ledger.register_nym.return_value = (
                True,
                {"signed_txn": {"...": "..."}},
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.register_ledger_nym(self.request)

    async def test_register_nym_create_transaction_for_endorser_not_found_x(self):
        self.request.query = {
            "did": "a_test_did",
            "verkey": "a_test_verkey",
            "alias": "did_alias",
            "role": "ENDORSER",
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.side_effect = test_module.StorageNotFoundError()
            self.ledger.register_nym.return_value = (
                True,
                {"signed_txn": {"...": "..."}},
            )

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.register_ledger_nym(self.request)

    async def test_register_nym_create_transaction_for_endorser_base_model_x(self):
        self.request.query = {
            "did": "a_test_did",
            "verkey": "a_test_verkey",
            "alias": "did_alias",
            "role": "ENDORSER",
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.side_effect = test_module.BaseModelError()
            self.ledger.register_nym.return_value = (
                True,
                {"signed_txn": {"...": "..."}},
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.register_ledger_nym(self.request)

    async def test_register_nym_create_transaction_for_endorser_no_endorser_info_x(
        self,
    ):
        self.request.query = {
            "did": "a_test_did",
            "verkey": "a_test_verkey",
            "alias": "did_alias",
            "role": "ENDORSER",
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(return_value=None)
            )
            self.ledger.register_nym.return_value = (
                True,
                {"signed_txn": {"...": "..."}},
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.register_ledger_nym(self.request)

    async def test_register_nym_create_transaction_for_endorser_no_endorser_did_x(self):
        self.request.query = {
            "did": "a_test_did",
            "verkey": "a_test_verkey",
            "alias": "did_alias",
            "role": "ENDORSER",
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "endorser_name": ("name"),
                    }
                )
            )
            self.ledger.register_nym.return_value = (
                True,
                {"signed_txn": {"...": "..."}},
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.register_ledger_nym(self.request)

    async def test_get_nym_role_a(self):
        self.mock_ledger_requests_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor,
            self.mock_ledger_requests_executor,
        )

        self.request.query = {"did": self.test_did}

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.ledger.get_nym_role.return_value = Role.USER
            result = await test_module.get_nym_role(self.request)
            json_response.assert_called_once_with({"role": "USER"})
            assert result is json_response.return_value

    async def test_get_nym_role_b(self):
        self.mock_ledger_requests_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=("test_ledger_id", self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor,
            self.mock_ledger_requests_executor,
        )

        self.request.query = {"did": self.test_did}

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.ledger.get_nym_role.return_value = Role.USER
            result = await test_module.get_nym_role(self.request)
            json_response.assert_called_once_with(
                {"ledger_id": "test_ledger_id", "role": "USER"}
            )
            assert result is json_response.return_value

    async def test_get_nym_role_multitenant(self):
        self.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        self.request.query = {"did": self.test_did}

        with (
            mock.patch.object(
                IndyLedgerRequestsExecutor,
                "get_ledger_for_identifier",
                mock.CoroutineMock(return_value=("test_ledger_id", self.ledger)),
            ),
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as json_response,
        ):
            self.ledger.get_nym_role.return_value = Role.USER
            result = await test_module.get_nym_role(self.request)
            json_response.assert_called_once_with(
                {"ledger_id": "test_ledger_id", "role": "USER"}
            )
            assert result is json_response.return_value

    async def test_get_nym_role_bad_request(self):
        self.request.query = {"no": "did"}
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.get_nym_role(self.request)

    async def test_get_nym_role_ledger_txn_error(self):
        self.mock_ledger_requests_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=("test_ledger_id", self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor,
            self.mock_ledger_requests_executor,
        )
        self.request.query = {"did": self.test_did}
        self.ledger.get_nym_role.side_effect = test_module.LedgerTransactionError(
            "Error in building get-nym request"
        )
        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.get_nym_role(self.request)

    async def test_get_nym_role_bad_ledger_req(self):
        self.mock_ledger_requests_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=("test_ledger_id", self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor,
            self.mock_ledger_requests_executor,
        )
        self.request.query = {"did": self.test_did}
        self.ledger.get_nym_role.side_effect = test_module.BadLedgerRequestError(
            "No such public DID"
        )
        with self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.get_nym_role(self.request)

    async def test_get_nym_role_ledger_error(self):
        self.mock_ledger_requests_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor,
            self.mock_ledger_requests_executor,
        )
        self.request.query = {"did": self.test_did}
        self.ledger.get_nym_role.side_effect = test_module.LedgerError("Error")
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.get_nym_role(self.request)

    async def test_rotate_public_did_keypair(self):
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.ledger.rotate_public_did_keypair = mock.CoroutineMock()

            await test_module.rotate_public_did_keypair(self.request)
            json_response.assert_called_once_with({})

    async def test_rotate_public_did_keypair_public_wallet_x(self):
        with mock.patch.object(test_module.web, "json_response", mock.Mock()):
            self.ledger.rotate_public_did_keypair = mock.CoroutineMock(
                side_effect=test_module.WalletError("Exception")
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.rotate_public_did_keypair(self.request)

    async def test_get_taa(self):
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.ledger.get_txn_author_agreement.return_value = {"taa_required": False}
            self.ledger.get_latest_txn_author_acceptance.return_value = None
            result = await test_module.ledger_get_taa(self.request)
            json_response.assert_called_once_with(
                {"result": {"taa_accepted": None, "taa_required": False}}
            )
            assert result is json_response.return_value

    async def test_get_taa_required(self):
        accepted = {
            "mechanism": "dummy",
            "time": 1234567890,
        }
        taa_info = {"taa_required": True}

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.ledger.get_txn_author_agreement.return_value = taa_info
            self.ledger.get_latest_txn_author_acceptance.return_value = accepted
            result = await test_module.ledger_get_taa(self.request)
            taa_info["taa_accepted"] = accepted
            json_response.assert_called_once_with({"result": taa_info})
            assert result is json_response.return_value

    async def test_get_taa_x(self):
        self.ledger.get_txn_author_agreement.side_effect = test_module.LedgerError()

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.ledger_get_taa(self.request)

    async def test_taa_accept_not_required(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "version": "version",
                "text": "text",
                "mechanism": "mechanism",
            }
        )

        with self.assertRaises(test_module.web.HTTPBadRequest):
            self.ledger.get_txn_author_agreement.return_value = {"taa_required": False}
            await test_module.ledger_accept_taa(self.request)

    async def test_accept_taa(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "version": "version",
                "text": "text",
                "mechanism": "mechanism",
            }
        )

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.ledger.get_txn_author_agreement.return_value = {
                "taa_record": {"text": "text"},
                "taa_required": True,
            }
            result = await test_module.ledger_accept_taa(self.request)
            json_response.assert_called_once_with({})
            self.ledger.accept_txn_author_agreement.assert_awaited_once_with(
                {
                    "version": "version",
                    "text": "text",
                    "digest": self.ledger.taa_digest.return_value,
                },
                "mechanism",
            )
            assert result is json_response.return_value

    async def test_accept_taa_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "version": "version",
                "text": "text",
                "mechanism": "mechanism",
            }
        )
        self.ledger.get_txn_author_agreement.return_value = {
            "taa_record": {"text": "text"},
            "taa_required": True,
        }
        self.ledger.accept_txn_author_agreement.side_effect = test_module.StorageError()
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.ledger_accept_taa(self.request)

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]

    async def test_get_write_ledger(self):
        mock_manager = mock.MagicMock(BaseMultipleLedgerManager, autospec=True)
        mock_manager.get_ledger_id_by_ledger_pool_name = mock.CoroutineMock(
            return_value="test_ledger_id"
        )
        self.profile.context.injector.bind_instance(
            BaseMultipleLedgerManager, mock_manager
        )
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            result = await test_module.get_write_ledger(self.request)
            json_response.assert_called_once_with(
                {
                    "ledger_id": "test_ledger_id",
                }
            )
            assert result is json_response.return_value

    async def test_get_write_ledger_single_ledger(self):
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            result = await test_module.get_write_ledger(self.request)
            json_response.assert_called_once_with(
                {
                    "ledger_id": "default",
                }
            )
            assert result is json_response.return_value

    async def test_get_ledger_config(self):
        mock_manager = mock.MagicMock(BaseMultipleLedgerManager, autospec=True)
        mock_manager.get_prod_ledgers = mock.CoroutineMock(
            return_value={
                "test_1": mock.MagicMock(),
                "test_2": mock.MagicMock(),
                "test_5": mock.MagicMock(),
            }
        )
        mock_manager.get_nonprod_ledgers = mock.CoroutineMock(
            return_value={
                "test_3": mock.MagicMock(),
                "test_4": mock.MagicMock(),
            }
        )
        self.profile.context.injector.bind_instance(
            BaseMultipleLedgerManager, mock_manager
        )

        self.context.settings["ledger.ledger_config_list"] = [
            {"id": "test_1", "genesis_transactions": "..."},
            {"id": "test_2", "genesis_transactions": "..."},
            {"id": "test_3", "genesis_transactions": "..."},
            {"id": "test_4", "genesis_transactions": "..."},
        ]
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            result = await test_module.get_ledger_config(self.request)
            json_response.assert_called_once_with(
                {
                    "production_ledgers": [
                        {"id": "test_1"},
                        {"id": "test_2"},
                        {
                            "id": "test_5",
                            "desc": "ledger configured outside --genesis-transactions-list",
                        },
                    ],
                    "non_production_ledgers": [{"id": "test_3"}, {"id": "test_4"}],
                }
            )
            assert result is json_response.return_value

    async def test_get_ledger_config_x(self):
        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.get_ledger_config(self.request)

    async def test_get_ledger_config_structure(self):
        """Test the structure of the ledger config response."""
        mock_manager = mock.MagicMock(BaseMultipleLedgerManager, autospec=True)
        mock_manager.get_prod_ledgers = mock.CoroutineMock(return_value={"test_1": None})
        mock_manager.get_nonprod_ledgers = mock.CoroutineMock(
            return_value={"test_2": None}
        )
        self.profile.context.injector.bind_instance(
            BaseMultipleLedgerManager, mock_manager
        )

        self.context.settings["ledger.ledger_config_list"] = [
            {
                "id": "test_1",
                "is_production": True,
                "is_write": True,
                "keepalive": 5,
                "read_only": False,
                "pool_name": "test_pool",
                "socks_proxy": None,
            },
            {
                "id": "test_2",
                "is_production": False,
                "is_write": False,
                "keepalive": 10,
                "read_only": True,
                "pool_name": "non_prod_pool",
                "socks_proxy": None,
            },
        ]

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            await test_module.get_ledger_config(self.request)

            response_data = json_response.call_args[0][0]
            assert "production_ledgers" in response_data
            assert "non_production_ledgers" in response_data

            prod_ledger = response_data["production_ledgers"][0]
            assert prod_ledger == {
                "id": "test_1",
                "is_production": True,
                "is_write": True,
                "keepalive": 5,
                "read_only": False,
                "pool_name": "test_pool",
                "socks_proxy": None,
            }

            non_prod_ledger = response_data["non_production_ledgers"][0]
            assert non_prod_ledger == {
                "id": "test_2",
                "is_production": False,
                "is_write": False,
                "keepalive": 10,
                "read_only": True,
                "pool_name": "non_prod_pool",
                "socks_proxy": None,
            }

    async def test_ledger_config_schema_validation(self):
        """Test schema validation for required fields."""
        schema = LedgerConfigInstanceSchema()

        minimal_data = {
            "is_production": True,
            "is_write": False,
            "keepalive": 5,
            "read_only": False,
        }
        loaded = schema.load(minimal_data)
        assert loaded.pool_name == loaded.id
        assert loaded.is_write is False

        with pytest.raises(ValidationError) as exc:
            schema.load({"is_production": "not_bool"})
        assert "is_production" in exc.value.messages

    async def test_ledger_config_id_generation(self):
        """Test automatic ID generation when missing."""
        schema = LedgerConfigInstanceSchema()

        data = {
            "is_production": True,
            "is_write": False,  # Add required fields
            "keepalive": 5,
            "read_only": False,
        }
        loaded = schema.load(data)
        assert uuid.UUID(loaded.id, version=4)

        explicit_id = str(uuid4())
        loaded = schema.load({"id": explicit_id, "is_production": True})
        assert loaded.id == explicit_id

    async def test_empty_ledger_lists(self):
        schema = LedgerConfigListSchema()
        empty_data = {"production_ledgers": [], "non_production_ledgers": []}
        loaded = schema.load(empty_data)
        assert loaded == empty_data

    # Multiple Ledgers Configured
    async def test_get_write_ledgers_multiple(self):
        # Mock the multiple ledger manager
        mock_manager = mock.MagicMock(BaseMultipleLedgerManager)
        mock_manager.get_write_ledgers = mock.CoroutineMock(
            return_value=["ledger1", "ledger2", "ledger3"]
        )
        self.profile.context.injector.bind_instance(
            BaseMultipleLedgerManager, mock_manager
        )

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            result = await test_module.get_write_ledgers(self.request)

            # Assert the response matches the expected structure
            json_response.assert_called_once_with(
                {"write_ledgers": ["ledger1", "ledger2", "ledger3"]}
            )
            assert result is json_response.return_value

    # Single Ledger (No Multi-Ledger Manager)
    async def test_get_write_ledgers_single(self):
        # Ensure no multi-ledger manager is bound
        self.profile.context.injector.clear_binding(BaseMultipleLedgerManager)

        result = await test_module.get_write_ledgers(self.request)

        # Extract the JSON body from the response
        response_body = result.text
        response_body = json.loads(response_body)

        # Assert the response is correct
        self.assertEqual(response_body, {"write_ledgers": ["default"]})

    # Schema Validation
    async def test_get_write_ledgers_schema(self):
        # Mock the multiple ledger manager
        mock_manager = mock.MagicMock(BaseMultipleLedgerManager)
        mock_manager.get_write_ledgers = mock.CoroutineMock(
            return_value=["ledger1", "ledger2"]
        )
        self.profile.context.injector.bind_instance(
            BaseMultipleLedgerManager, mock_manager
        )

        response = await test_module.get_write_ledgers(self.request)

        # Validate against the schema
        schema = ConfigurableWriteLedgersSchema()
        data = json.loads(response.body)
        validated = schema.validate(data)
        assert validated == {}



"""Validates State Proof."""

import json
from collections import OrderedDict

from rlp import DecodingError
from rlp import decode as rlp_decode
from rlp import encode as rlp_encode

from .constants import (
    BLANK_NODE,
    NODE_TYPE_BLANK,
    NODE_TYPE_BRANCH,
    NODE_TYPE_EXTENSION,
    NODE_TYPE_LEAF,
)
from .utils import NIBBLE_TERMINATOR, sha3_256, unpack_to_nibbles


class SubTrie:
    """Utility class for SubTrie and State Proof validation."""

    def __init__(self, root_hash=None):
        """MPT SubTrie dictionary like interface."""
        self._subtrie = OrderedDict()
        self.set_root_hash(root_hash)

    @property
    def root_hash(self):
        """Return 32 bytes string."""
        return self._root_hash

    @root_hash.setter
    def root_hash(self, value):
        self.set_root_hash(value)

    def set_root_hash(self, root_hash=None):
        """."""
        self._root_hash = root_hash

    @staticmethod
    def _get_node_type(node):
        if node == BLANK_NODE:
            return NODE_TYPE_BLANK
        if len(node) == 2:
            nibbles = unpack_to_nibbles(node[0])
            has_terminator = nibbles and nibbles[-1] == NIBBLE_TERMINATOR
            return NODE_TYPE_LEAF if has_terminator else NODE_TYPE_EXTENSION
        if len(node) == 17:
            return NODE_TYPE_BRANCH

    @staticmethod
    async def verify_spv_proof(expected_value, proof_nodes, serialized=True):
        """Verify State Proof."""
        try:
            if serialized:
                proof_nodes = rlp_decode(proof_nodes)
            new_trie = await SubTrie.get_new_trie_with_proof_nodes(proof_nodes)
            expected_value = json.loads(expected_value)
            for encoded_node in list(new_trie._subtrie.values()):
                try:
                    decoded_node = rlp_decode(encoded_node)
                    # branch node
                    if SubTrie._get_node_type(decoded_node) == NODE_TYPE_BRANCH:
                        if (
                            json.loads(rlp_decode(decoded_node[-1])[0].decode("utf-8"))
                        ) == expected_value:
                            return True
                    # leaf or extension node
                    if SubTrie._get_node_type(decoded_node) == NODE_TYPE_LEAF:
                        if (
                            json.loads(rlp_decode(decoded_node[1])[0].decode("utf-8"))
                        ) == expected_value:
                            return True
                except DecodingError:
                    continue
            return False
        except Exception:
            return False

    @staticmethod
    async def get_new_trie_with_proof_nodes(proof_nodes):
        """Return SubTrie created from proof_nodes."""
        new_trie = SubTrie()
        for node in proof_nodes:
            R = rlp_encode(node)
            H = sha3_256(R)
            new_trie._subtrie[H] = R
        return new_trie



"""Event tracing."""

import datetime
import json
import logging
import time
from typing import Optional

import requests
from marshmallow import fields

from ..messaging.agent_message import AgentMessage
from ..messaging.decorators.trace_decorator import (
    TRACE_LOG_TARGET,
    TRACE_MESSAGE_TARGET,
    TraceReport,
)
from ..messaging.models.base_record import BaseExchangeRecord
from ..messaging.models.openapi import OpenAPISchema
from ..transport.inbound.message import InboundMessage
from ..transport.outbound.message import OutboundMessage

LOGGER = logging.getLogger(__name__)
DT_FMT = "%Y-%m-%d %H:%M:%S.%f%z"


class AdminAPIMessageTracingSchema(OpenAPISchema):
    """Request/result schema including agent message tracing.

    This is to be used as a superclass for aca-py admin input/output
    messages that need to support tracing.
    """

    trace = fields.Boolean(
        required=False,
        dump_default=False,
        metadata={
            "description": "Record trace information, based on agent configuration"
        },
    )


def get_timer() -> float:
    """Return a timer."""
    return time.perf_counter()


def tracing_enabled(context, message) -> bool:
    """Determine whether to log trace messages or not."""
    # check if tracing is explicitly on
    if context.get("trace.enabled"):
        return True

    if message:
        if isinstance(message, AgentMessage):
            # if there is a trace decorator on the messages then continue to trace
            if message._trace:
                return True
        elif isinstance(message, BaseExchangeRecord):
            if message.trace:
                return True
        elif isinstance(message, dict):
            # if there is a trace decorator on the messages then continue to trace
            if message.get("~trace"):
                return True
            if message.get("trace"):
                return message.get("trace")
        elif isinstance(message, str):
            if "~trace" in message:
                return True
            if "trace" in message:
                msg = json.loads(message)
                return msg.get("trace")
        elif isinstance(message, OutboundMessage):
            if message.payload and isinstance(message.payload, AgentMessage):
                if message.payload._trace:
                    return True
            elif message.payload and isinstance(message.payload, dict):
                if message.payload.get("~trace") or message.payload.get("trace"):
                    return True
            elif message.payload and isinstance(message.payload, str):
                if "trace" in message.payload:  # includes "~trace" in message.payload
                    return True

    # default off
    return False


def decode_inbound_message(message):
    """Return bundled message if appropriate."""

    if message and isinstance(message, OutboundMessage):
        if message.payload and isinstance(message.payload, AgentMessage):
            return message.payload
        elif message.payload and isinstance(message.payload, dict):
            return message.payload
        elif message.payload and isinstance(message.payload, str):
            try:
                return json.loads(message.payload)
            except Exception:
                pass
    elif message and isinstance(message, str):
        try:
            return json.loads(message)
        except Exception:
            pass

    # default is the provided message
    return message


def trace_event(
    context,
    message,
    handler: Optional[str] = None,
    outcome: Optional[str] = None,
    perf_counter: Optional[float] = None,
    force_trace: bool = False,
    raise_errors: bool = False,
) -> float:
    """Log a trace event to a configured target.

    Args:
        context: The application context, attributes of interest are:
            context["trace.enabled"]: True if we are logging events
            context["trace.target"]: Trace target ("log", "message" or an http endpoint)
            context["trace.tag"]: Tag to be included in trace output
        message: the current message, can be an AgentMessage,
            InboundMessage, OutboundMessage or Exchange record
        handler (optional): The handler name for the trace event. If not provided,
            it defaults to "aca-py.agent".
        outcome (optional): The outcome of the trace event.
        perf_counter (optional): The performance counter value for the trace event.
        force_trace (optional): If True, forces the trace event to be logged even if
            tracing is not enabled.
        raise_errors (optional): If True, raises an exception if there is an error
            logging the trace event.

    Returns:
        float: The value of the performance counter.

    Raises:
        Exception: If there is an error logging the trace event and `raise_errors` is
            True.

    """

    ret = time.perf_counter()

    if force_trace or tracing_enabled(context, message):
        message = decode_inbound_message(message)

        # build the event to log
        # TODO check instance type of message to determine how to
        # get message and thread id's
        if not handler:
            if context and context.get("trace.label"):
                handler = context.get("trace.label")
            else:
                handler = "aca-py.agent"
        msg_id = ""
        thread_id = ""
        msg_type = ""
        if message and isinstance(message, AgentMessage):
            msg_id = str(message._id)
            if message._thread and message._thread.thid:
                thread_id = str(message._thread.thid)
            else:
                thread_id = msg_id
            msg_type = str(message._type)
        elif message and isinstance(message, InboundMessage):
            # TODO not sure if we can log an InboundMessage before it's "handled"
            msg_id = str(message.session_id) if message.session_id else "N/A"
            thread_id = str(message.session_id) if message.session_id else "N/A"
            msg_type = str(message.__class__.__name__)
        elif message and isinstance(message, OutboundMessage):
            msg_id = str(message.reply_thread_id) if message.reply_thread_id else "N/A"
            thread_id = msg_id
            msg_type = str(message.__class__.__name__)
        elif message and isinstance(message, dict):
            msg_id = str(message["@id"]) if message.get("@id") else "N/A"
            if message.get("~thread") and message["~thread"].get("thid"):
                thread_id = str(message["~thread"]["thid"])
            elif message.get("thread_id"):
                thread_id = str(message["thread_id"])
            else:
                thread_id = msg_id
            if message.get("@type"):
                msg_type = str(message["@type"])
            else:
                if message.get("~thread"):
                    msg_type = "dict:Message"
                elif message.get("thread_id"):
                    msg_type = "dict:Exchange"
                else:
                    msg_type = "dict"
        elif isinstance(message, BaseExchangeRecord):
            msg_id = "N/A"
            thread_id = str(message.thread_id)
            msg_type = str(message.__class__.__name__)
        else:
            msg_id = "N/A"
            thread_id = "N/A"
            msg_type = str(message.__class__.__name__)
        ep_time = time.time()
        str_time = datetime.datetime.fromtimestamp(
            ep_time, datetime.timezone.utc
        ).strftime(DT_FMT)
        event = {
            "msg_id": msg_id,
            "thread_id": thread_id if thread_id else msg_id,
            "traced_type": msg_type,
            "timestamp": ep_time,
            "str_time": str_time,
            "handler": str(handler),
            "elapsed_milli": int(1000 * (ret - perf_counter)) if perf_counter else 0,
            "outcome": str(outcome),
        }
        event_str = json.dumps(event)

        try:
            # check our target - if we get this far we know we are logging the event
            if context["trace.target"] == TRACE_MESSAGE_TARGET and isinstance(
                message, AgentMessage
            ):
                # add a trace report to the existing message
                trace_report = TraceReport(
                    msg_id=event["msg_id"],
                    thread_id=event["thread_id"],
                    traced_type=event["traced_type"],
                    timestamp=event["timestamp"],
                    str_time=event["str_time"],
                    handler=event["handler"],
                    elapsed_milli=event["elapsed_milli"],
                    outcome=event["outcome"],
                )
                message.add_trace_report(trace_report)
            elif context["trace.target"] == TRACE_LOG_TARGET:
                # write to standard log file
                LOGGER.setLevel(logging.INFO)
                LOGGER.info(" %s %s", context["trace.tag"], event_str)
            else:
                # should be an http endpoint
                _ = requests.post(
                    context["trace.target"]
                    + (context["trace.tag"] if context["trace.tag"] else ""),
                    data=event_str,
                    headers={"Content-Type": "application/json"},
                )
        except Exception as e:
            if raise_errors:
                raise
            LOGGER.error(
                "Error logging trace target: %s tag: %s event: %s",
                context.get("trace.target"),
                context.get("trace.tag"),
                event_str,
            )
            LOGGER.exception(e)

    else:
        # trace is not enabled so just return
        pass

    return ret



"""Marshmallow bindings for indy proofs."""

from typing import Mapping, Optional, Sequence

from marshmallow import EXCLUDE, fields, validate

from ...messaging.models.base import BaseModel, BaseModelSchema
from ...messaging.valid import (
    INDY_CRED_DEF_ID_EXAMPLE,
    INDY_CRED_DEF_ID_VALIDATE,
    INDY_REV_REG_ID_EXAMPLE,
    INDY_REV_REG_ID_VALIDATE,
    INDY_SCHEMA_ID_EXAMPLE,
    INDY_SCHEMA_ID_VALIDATE,
    INT_EPOCH_EXAMPLE,
    INT_EPOCH_VALIDATE,
    NUM_STR_ANY_EXAMPLE,
    NUM_STR_ANY_VALIDATE,
    NUM_STR_WHOLE_EXAMPLE,
    NUM_STR_WHOLE_VALIDATE,
)
from ...utils.tracing import AdminAPIMessageTracingSchema
from .predicate import Predicate
from .requested_creds import (
    IndyRequestedCredsRequestedAttrSchema,
    IndyRequestedCredsRequestedPredSchema,
)


class IndyEQProof(BaseModel):
    """Equality proof for indy primary proof."""

    class Meta:
        """Equality proof metadata."""

        schema_class = "IndyEQProofMeta"

    def __init__(
        self,
        revealed_attrs: Mapping[str, str] = None,
        a_prime: Optional[str] = None,
        e: Optional[str] = None,
        v: Optional[str] = None,
        m: Mapping[str, str] = None,
        m2: Optional[str] = None,
        **kwargs,
    ):
        """Initialize equality proof object."""
        super().__init__(**kwargs)
        self.revealed_attrs = revealed_attrs
        self.a_prime = a_prime
        self.e = e
        self.v = v
        self.m = m
        self.m2 = m2


class IndyEQProofSchema(BaseModelSchema):
    """Indy equality proof schema."""

    class Meta:
        """Indy equality proof metadata."""

        model_class = IndyEQProof
        unknown = EXCLUDE

    revealed_attrs = fields.Dict(
        keys=fields.Str(metadata={"example": "preference"}),
        values=fields.Str(
            validate=NUM_STR_ANY_VALIDATE, metadata={"example": NUM_STR_ANY_EXAMPLE}
        ),
    )
    a_prime = fields.Str(
        validate=NUM_STR_WHOLE_VALIDATE, metadata={"example": NUM_STR_WHOLE_EXAMPLE}
    )
    e = fields.Str(
        validate=NUM_STR_WHOLE_VALIDATE, metadata={"example": NUM_STR_WHOLE_EXAMPLE}
    )
    v = fields.Str(
        validate=NUM_STR_WHOLE_VALIDATE, metadata={"example": NUM_STR_WHOLE_EXAMPLE}
    )
    m = fields.Dict(
        keys=fields.Str(metadata={"example": "master_secret"}),
        values=fields.Str(
            validate=NUM_STR_WHOLE_VALIDATE, metadata={"example": NUM_STR_WHOLE_EXAMPLE}
        ),
    )
    m2 = fields.Str(
        validate=NUM_STR_WHOLE_VALIDATE, metadata={"example": NUM_STR_WHOLE_EXAMPLE}
    )


class IndyGEProofPred(BaseModel):
    """Indy GE proof predicate."""

    class Meta:
        """Indy GE proof predicate metadata."""

        schema_class = "IndyGEProofPredSchema"

    def __init__(
        self,
        attr_name: Optional[str] = None,
        p_type: Optional[str] = None,
        value: Optional[int] = None,
        **kwargs,
    ):
        """Initialize indy GE proof predicate."""
        super().__init__(**kwargs)
        self.attr_name = attr_name
        self.p_type = p_type
        self.value = value


class IndyGEProofPredSchema(BaseModelSchema):
    """Indy GE proof predicate schema."""

    class Meta:
        """Indy GE proof predicate metadata."""

        model_class = IndyGEProofPred
        unknown = EXCLUDE

    attr_name = fields.Str(metadata={"description": "Attribute name, indy-canonicalized"})
    p_type = fields.Str(
        validate=validate.OneOf([p.fortran for p in Predicate]),
        metadata={"description": "Predicate type"},
    )
    value = fields.Integer(
        metadata={"strict": True, "description": "Predicate threshold value"}
    )


class IndyGEProof(BaseModel):
    """Greater-than-or-equal-to proof for indy primary proof."""

    class Meta:
        """GE proof metadata."""

        schema_class = "IndyGEProofMeta"

    def __init__(
        self,
        u: Mapping[str, str] = None,
        r: Mapping[str, str] = None,
        mj: Optional[str] = None,
        alpha: Optional[str] = None,
        t: Mapping[str, str] = None,
        predicate: Optional[IndyGEProofPred] = None,
        **kwargs,
    ):
        """Initialize GE proof object."""
        super().__init__(**kwargs)
        self.u = u
        self.r = r
        self.mj = mj
        self.alpha = alpha
        self.t = t
        self.predicate = predicate


class IndyGEProofSchema(BaseModelSchema):
    """Indy GE proof schema."""

    class Meta:
        """Indy GE proof schema metadata."""

        model_class = IndyGEProof
        unknown = EXCLUDE

    u = fields.Dict(
        keys=fields.Str(),
        values=fields.Str(
            validate=NUM_STR_WHOLE_VALIDATE, metadata={"example": NUM_STR_WHOLE_EXAMPLE}
        ),
    )
    r = fields.Dict(
        keys=fields.Str(),
        values=fields.Str(
            validate=NUM_STR_WHOLE_VALIDATE, metadata={"example": NUM_STR_WHOLE_EXAMPLE}
        ),
    )
    mj = fields.Str(
        validate=NUM_STR_WHOLE_VALIDATE, metadata={"example": NUM_STR_WHOLE_EXAMPLE}
    )
    alpha = fields.Str(
        validate=NUM_STR_WHOLE_VALIDATE, metadata={"example": NUM_STR_WHOLE_EXAMPLE}
    )
    t = fields.Dict(
        keys=fields.Str(),
        values=fields.Str(
            validate=NUM_STR_WHOLE_VALIDATE, metadata={"example": NUM_STR_WHOLE_EXAMPLE}
        ),
    )
    predicate = fields.Nested(IndyGEProofPredSchema)


class IndyPrimaryProof(BaseModel):
    """Indy primary proof."""

    class Meta:
        """Indy primary proof metadata."""

        schema_class = "IndyPrimaryProofSchema"

    def __init__(
        self,
        eq_proof: Optional[IndyEQProof] = None,
        ge_proofs: Sequence[IndyGEProof] = None,
        **kwargs,
    ):
        """Initialize indy primary proof."""
        super().__init__(**kwargs)
        self.eq_proof = eq_proof
        self.ge_proofs = ge_proofs


class IndyPrimaryProofSchema(BaseModelSchema):
    """Indy primary proof schema."""

    class Meta:
        """Indy primary proof schema metadata."""

        model_class = IndyPrimaryProof
        unknown = EXCLUDE

    eq_proof = fields.Nested(
        IndyEQProofSchema,
        allow_none=True,
        metadata={"description": "Indy equality proof"},
    )
    ge_proofs = fields.Nested(
        IndyGEProofSchema,
        many=True,
        allow_none=True,
        metadata={"description": "Indy GE proofs"},
    )


class IndyNonRevocProof(BaseModel):
    """Indy non-revocation proof."""

    class Meta:
        """Indy non-revocation proof metadata."""

        schema_class = "IndyNonRevocProofSchema"

    def __init__(
        self,
        x_list: Optional[Mapping] = None,
        c_list: Optional[Mapping] = None,
        **kwargs,
    ):
        """Initialize indy non-revocation proof."""
        super().__init__(**kwargs)
        self.x_list = x_list
        self.c_list = c_list


class IndyNonRevocProofSchema(BaseModelSchema):
    """Indy non-revocation proof schema."""

    class Meta:
        """Indy non-revocation proof schema metadata."""

        model_class = IndyNonRevocProof
        unknown = EXCLUDE

    x_list = fields.Dict(keys=fields.Str(), values=fields.Str())
    c_list = fields.Dict(keys=fields.Str(), values=fields.Str())


class IndyProofProofProofsProof(BaseModel):
    """Indy proof.proof.proofs constituent proof."""

    class Meta:
        """Indy proof.proof.proofs constituent proof schema."""

        schema_class = "IndyProofProofProofsProofSchema"

    def __init__(
        self,
        primary_proof: Optional[IndyPrimaryProof] = None,
        non_revoc_proof: Optional[IndyNonRevocProof] = None,
        **kwargs,
    ):
        """Initialize proof.proof.proofs constituent proof."""
        super().__init__(**kwargs)
        self.primary_proof = primary_proof
        self.non_revoc_proof = non_revoc_proof


class IndyProofProofProofsProofSchema(BaseModelSchema):
    """Indy proof.proof.proofs constituent proof schema."""

    class Meta:
        """Indy proof.proof.proofs constituent proof schema metadata."""

        model_class = IndyProofProofProofsProof
        unknown = EXCLUDE

    primary_proof = fields.Nested(
        IndyPrimaryProofSchema, metadata={"description": "Indy primary proof"}
    )
    non_revoc_proof = fields.Nested(
        IndyNonRevocProofSchema,
        allow_none=True,
        metadata={"description": "Indy non-revocation proof"},
    )


class IndyProofProofAggregatedProof(BaseModel):
    """Indy proof.proof aggregated proof."""

    class Meta:
        """Indy proof.proof aggregated proof metadata."""

        schema_class = "IndyProofProofAggregatedProofSchema"

    def __init__(
        self,
        c_hash: Optional[str] = None,
        c_list: Sequence[Sequence[int]] = None,
        **kwargs,
    ):
        """Initialize indy proof.proof agreggated proof."""
        super().__init__(**kwargs)
        self.c_hash = c_hash
        self.c_list = c_list


class IndyProofProofAggregatedProofSchema(BaseModelSchema):
    """Indy proof.proof aggregated proof schema."""

    class Meta:
        """Indy proof.proof aggregated proof schema metadata."""

        model_class = IndyProofProofAggregatedProof
        unknown = EXCLUDE

    c_hash = fields.Str(metadata={"description": "c_hash value"})
    c_list = fields.List(
        fields.List(fields.Int(metadata={"strict": True})),
        metadata={"description": "c_list value"},
    )


class IndyProofProof(BaseModel):
    """Indy proof.proof content."""

    class Meta:
        """Indy proof.proof content metadata."""

        schema_class = "IndyProofProofSchema"

    def __init__(
        self,
        proofs: Sequence[IndyProofProofProofsProof] = None,
        aggregated_proof: Optional[IndyProofProofAggregatedProof] = None,
        **kwargs,
    ):
        """Initialize indy proof.proof content."""
        super().__init__(**kwargs)
        self.proofs = proofs
        self.aggregated_proof = aggregated_proof


class IndyProofProofSchema(BaseModelSchema):
    """Indy proof.proof content schema."""

    class Meta:
        """Indy proof.proof content schema metadata."""

        model_class = IndyProofProof
        unknown = EXCLUDE

    proofs = fields.Nested(
        IndyProofProofProofsProofSchema,
        many=True,
        metadata={"description": "Indy proof proofs"},
    )
    aggregated_proof = fields.Nested(
        IndyProofProofAggregatedProofSchema,
        metadata={"description": "Indy proof aggregated proof"},
    )


class RawEncoded(BaseModel):
    """Raw and encoded attribute values."""

    class Meta:
        """Raw and encoded attribute values metadata."""

        schema_class = "RawEncodedSchema"

    def __init__(
        self,
        raw: Optional[str] = None,
        encoded: Optional[str] = None,
        **kwargs,
    ):
        """Initialize raw and encoded attribute values."""
        super().__init__(**kwargs)
        self.raw = raw
        self.encoded = encoded


class RawEncodedSchema(BaseModelSchema):
    """Raw and encoded attribute values schema."""

    class Meta:
        """Raw and encoded attribute values schema metadata."""

        model_class = RawEncoded
        unknown = EXCLUDE

    raw = fields.Str(metadata={"description": "Raw value"})
    encoded = fields.Str(
        validate=NUM_STR_ANY_VALIDATE,
        metadata={"description": "Encoded value", "example": NUM_STR_ANY_EXAMPLE},
    )


class IndyProofRequestedProofRevealedAttr(RawEncoded):
    """Indy proof requested proof revealed attr."""

    class Meta:
        """Indy proof requested proof revealed attr metadata."""

        schema_class = "IndyProofRequestedProofRevealedAttrSchema"

    def __init__(
        self,
        sub_proof_index: Optional[int] = None,
        **kwargs,
    ):
        """Initialize indy proof requested proof revealed attr."""
        super().__init__(**kwargs)
        self.sub_proof_index = sub_proof_index


class IndyProofRequestedProofRevealedAttrSchema(RawEncodedSchema):
    """Indy proof requested proof revealed attr schema."""

    class Meta:
        """Indy proof requested proof revealed attr schema metadata."""

        model_class = IndyProofRequestedProofRevealedAttr
        unknown = EXCLUDE

    sub_proof_index = fields.Int(
        metadata={"strict": True, "description": "Sub-proof index"}
    )


class IndyProofRequestedProofRevealedAttrGroup(BaseModel):
    """Indy proof requested proof revealed attr group."""

    class Meta:
        """Indy proof requested proof revealed attr group metadata."""

        schema_class = "IndyProofRequestedProofRevealedAttrGroupSchema"

    def __init__(
        self,
        sub_proof_index: Optional[int] = None,
        values: Mapping[str, RawEncoded] = None,
        **kwargs,
    ):
        """Initialize indy proof requested proof revealed attr."""
        super().__init__(**kwargs)
        self.sub_proof_index = sub_proof_index
        self.values = values


class IndyProofRequestedProofRevealedAttrGroupSchema(BaseModelSchema):
    """Indy proof requested proof revealed attr group schema."""

    class Meta:
        """Indy proof requested proof revealed attr group schema metadata."""

        model_class = IndyProofRequestedProofRevealedAttrGroup
        unknown = EXCLUDE

    sub_proof_index = fields.Int(
        metadata={"strict": True, "description": "Sub-proof index"}
    )
    values = fields.Dict(
        keys=fields.Str(),
        values=fields.Nested(RawEncodedSchema),
        metadata={
            "description": "Indy proof requested proof revealed attr groups group value"
        },
    )


class IndyProofRequestedProofPredicate(BaseModel):
    """Indy proof requested proof predicate."""

    class Meta:
        """Indy proof requested proof requested proof predicate metadata."""

        schema_class = "IndyProofRequestedProofPredicateSchema"

    def __init__(
        self,
        sub_proof_index: Optional[int] = None,
        **kwargs,
    ):
        """Initialize indy proof requested proof predicate."""
        super().__init__(**kwargs)
        self.sub_proof_index = sub_proof_index


class IndyProofRequestedProofPredicateSchema(BaseModelSchema):
    """Indy proof requested prrof predicate schema."""

    class Meta:
        """Indy proof requested proof requested proof predicate schema metadata."""

        model_class = IndyProofRequestedProofPredicate
        unknown = EXCLUDE

    sub_proof_index = fields.Int(
        metadata={"strict": True, "description": "Sub-proof index"}
    )


class IndyProofRequestedProof(BaseModel):
    """Indy proof.requested_proof content."""

    class Meta:
        """Indy proof.requested_proof content metadata."""

        schema_class = "IndyProofRequestedProofSchema"

    def __init__(
        self,
        revealed_attrs: Mapping[str, IndyProofRequestedProofRevealedAttr] = None,
        revealed_attr_groups: Mapping[
            str,
            IndyProofRequestedProofRevealedAttrGroup,
        ] = None,
        self_attested_attrs: Optional[Mapping] = None,
        unrevealed_attrs: Optional[Mapping] = None,
        predicates: Mapping[str, IndyProofRequestedProofPredicate] = None,
        **kwargs,
    ):
        """Initialize indy proof requested proof."""
        super().__init__(**kwargs)
        self.revealed_attrs = revealed_attrs
        self.revealed_attr_groups = revealed_attr_groups
        self.self_attested_attrs = self_attested_attrs
        self.unrevealed_attrs = unrevealed_attrs
        self.predicates = predicates


class IndyProofRequestedProofSchema(BaseModelSchema):
    """Indy proof requested proof schema."""

    class Meta:
        """Indy proof requested proof schema metadata."""

        model_class = IndyProofRequestedProof
        unknown = EXCLUDE

    revealed_attrs = fields.Dict(
        keys=fields.Str(),
        values=fields.Nested(IndyProofRequestedProofRevealedAttrSchema),
        allow_none=True,
        metadata={"description": "Proof requested proof revealed attributes"},
    )
    revealed_attr_groups = fields.Dict(
        keys=fields.Str(),
        values=fields.Nested(IndyProofRequestedProofRevealedAttrGroupSchema),
        allow_none=True,
        metadata={"description": "Proof requested proof revealed attribute groups"},
    )
    self_attested_attrs = fields.Dict(
        metadata={"description": "Proof requested proof self-attested attributes"}
    )
    unrevealed_attrs = fields.Dict(metadata={"description": "Unrevealed attributes"})
    predicates = fields.Dict(
        keys=fields.Str(),
        values=fields.Nested(IndyProofRequestedProofPredicateSchema),
        metadata={"description": "Proof requested proof predicates."},
    )


class IndyProofIdentifier(BaseModel):
    """Indy proof identifier."""

    class Meta:
        """Indy proof identifier metadata."""

        schema_class = "IndyProofIdentifierSchema"

    def __init__(
        self,
        schema_id: Optional[str] = None,
        cred_def_id: Optional[str] = None,
        rev_reg_id: Optional[str] = None,
        timestamp: Optional[int] = None,
        **kwargs,
    ):
        """Initialize indy proof identifier."""
        super().__init__(**kwargs)
        self.schema_id = schema_id
        self.cred_def_id = cred_def_id
        self.rev_reg_id = rev_reg_id
        self.timestamp = timestamp


class IndyProofIdentifierSchema(BaseModelSchema):
    """Indy proof identifier schema."""

    class Meta:
        """Indy proof identifier schema metadata."""

        model_class = IndyProofIdentifier
        unknown = EXCLUDE

    schema_id = fields.Str(
        validate=INDY_SCHEMA_ID_VALIDATE,
        metadata={
            "description": "Schema identifier",
            "example": INDY_SCHEMA_ID_EXAMPLE,
        },
    )
    cred_def_id = fields.Str(
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    rev_reg_id = fields.Str(
        allow_none=True,
        validate=INDY_REV_REG_ID_VALIDATE,
        metadata={
            "description": "Revocation registry identifier",
            "example": INDY_REV_REG_ID_EXAMPLE,
        },
    )
    timestamp = fields.Int(
        allow_none=True,
        validate=INT_EPOCH_VALIDATE,
        metadata={
            "strict": True,
            "description": "Timestamp epoch",
            "example": INT_EPOCH_EXAMPLE,
        },
    )


class IndyProof(BaseModel):
    """Indy proof."""

    class Meta:
        """Indy proof metadata."""

        schema_class = "IndyProofSchema"

    def __init__(
        self,
        proof: Optional[IndyProofProof] = None,
        requested_proof: Optional[IndyProofRequestedProof] = None,
        identifiers: Sequence[IndyProofIdentifier] = None,
        **kwargs,
    ):
        """Initialize indy proof."""
        super().__init__(**kwargs)
        self.proof = proof
        self.requested_proof = requested_proof
        self.identifiers = identifiers


class IndyProofSchema(BaseModelSchema):
    """Indy proof schema."""

    class Meta:
        """Indy proof schema metadata."""

        model_class = IndyProof
        unknown = EXCLUDE

    proof = fields.Nested(
        IndyProofProofSchema, metadata={"description": "Indy proof.proof content"}
    )
    requested_proof = fields.Nested(
        IndyProofRequestedProofSchema,
        metadata={"description": "Indy proof.requested_proof content"},
    )
    identifiers = fields.Nested(
        IndyProofIdentifierSchema,
        many=True,
        metadata={"description": "Indy proof.identifiers content"},
    )


class IndyPresSpecSchema(AdminAPIMessageTracingSchema):
    """Request schema for indy proof specification to send as presentation."""

    self_attested_attributes = fields.Dict(
        required=True,
        keys=fields.Str(metadata={"example": "attr_name"}),
        values=fields.Str(
            metadata={
                "example": "self_attested_value",
                "description": (
                    "Self-attested attribute values to use in requested-credentials"
                    " structure for proof construction"
                ),
            }
        ),
        metadata={"description": "Self-attested attributes to build into proof"},
    )
    requested_attributes = fields.Dict(
        required=True,
        keys=fields.Str(metadata={"example": "attr_referent"}),
        values=fields.Nested(IndyRequestedCredsRequestedAttrSchema),
        metadata={
            "description": (
                "Nested object mapping proof request attribute referents to"
                " requested-attribute specifiers"
            )
        },
    )
    requested_predicates = fields.Dict(
        required=True,
        keys=fields.Str(metadata={"example": "pred_referent"}),
        values=fields.Nested(IndyRequestedCredsRequestedPredSchema),
        metadata={
            "description": (
                "Nested object mapping proof request predicate referents to"
                " requested-predicate specifiers"
            )
        },
    )
    trace = fields.Bool(
        required=False,
        metadata={
            "description": "Whether to trace event (default false)",
            "example": False,
        },
    )



"""DIDComm V2 Agent message base class and schema."""

from uuid import uuid4

from .base_message import BaseMessage, DIDCommVersion


class V2AgentMessage(BaseMessage):
    """DIDComm V2 message base class."""

    def __init__(self, message, *args, **kwargs):
        """Initialize a v2 agent message instance."""
        super().__init__(*args, **kwargs)
        self.message = message
        self.msg_format = DIDCommVersion.v2
        if "id" not in self.message:
            self.message["id"] = str(uuid4())

    @property
    def _message_type(self):
        return self.message.get("type")

    @property
    def _type(self):
        return self.message.get("type")

    @property
    def _id(self):
        return self.message.get("id")

    @property
    def _thread_id(self):
        return self.message.get("thid", self.message.get("id"))

    def serialize(self, msg_format: DIDCommVersion = DIDCommVersion.v1) -> dict:
        """Serialize a message."""
        return self.message

    def deserialize(self, value: dict, msg_format: DIDCommVersion = DIDCommVersion.v2):
        """Deserialize a message."""
        self.message = value
        self.msg_format = msg_format

    @property
    def Handler(self):
        """Message handler."""
        return None



"""Verifiable Presentation model."""

from typing import List, Optional, Union

from marshmallow import INCLUDE, ValidationError, fields, post_dump

from ....messaging.models.base import BaseModel, BaseModelSchema
from ....messaging.valid import (
    CREDENTIAL_CONTEXT_VALIDATE,
    PRESENTATION_TYPE_EXAMPLE,
    PRESENTATION_TYPE_VALIDATE,
    DIDKey,
    StrOrDictField,
    Uri,
    UriOrDictField,
)
from ...ld_proofs.constants import (
    CREDENTIALS_CONTEXT_V1_URL,
    CREDENTIALS_CONTEXT_V2_URL,
    VERIFIABLE_PRESENTATION_TYPE,
)
from .linked_data_proof import LDProof, LinkedDataProofSchema


class VerifiablePresentation(BaseModel):
    """Single VerifiablePresentation object."""

    class Meta:
        """VerifiablePresentation metadata."""

        schema_class = "PresentationSchema"

    def __init__(
        self,
        context: Optional[List[Union[str, dict]]] = None,
        id: Optional[str] = None,
        type: Optional[List[str]] = None,
        holder: Optional[Union[dict, str]] = None,
        verifiable_credential: Optional[List[dict]] = None,
        proof: Optional[Union[dict, LDProof]] = None,
        **kwargs,
    ) -> None:
        """Initialize VerifiablePresentation."""
        self._context = context or [CREDENTIALS_CONTEXT_V1_URL]
        self._id = id
        self._holder = holder
        self._type = type or [VERIFIABLE_PRESENTATION_TYPE]
        self._verifiable_credential = verifiable_credential

        self._proof = proof

        self.extra = kwargs

    @property
    def context(self):
        """Getter for context."""
        return self._context

    @context.setter
    def context(self, context: List[Union[str, dict]]):
        """Setter for context.

        First item must be credentials v1 url
        """
        assert context[0] in [CREDENTIALS_CONTEXT_V1_URL, CREDENTIALS_CONTEXT_V2_URL]

        self._context = context

    def add_context(self, context: Union[str, dict]):
        """Add a context to this presentation."""
        self._context.append(context)

    @property
    def context_urls(self) -> List[str]:
        """Getter for context urls."""
        return [context for context in self.context if isinstance(context, str)]

    @property
    def type(self) -> List[str]:
        """Getter for type."""
        return self._type

    @type.setter
    def type(self, type: List[str]):
        """Setter for type.

        First item must be VerifiablePresentation
        """
        assert VERIFIABLE_PRESENTATION_TYPE in type

        self._type = type

    def add_type(self, type: str):
        """Add a type to this presentation."""
        self._type.append(type)

    @property
    def id(self):
        """Getter for id."""
        return self._id

    @id.setter
    def id(self, id: Union[str, None]):
        """Setter for id."""
        if id:
            uri_validator = Uri()
            uri_validator(id)

        self._id = id

    @property
    def holder_id(self) -> Optional[str]:
        """Getter for holder id."""
        if not self._holder:
            return None
        elif isinstance(self._holder, str):
            return self._holder

        return self._holder.get("id")

    @holder_id.setter
    def holder_id(self, holder_id: str):
        """Setter for holder id."""
        uri_validator = Uri()
        uri_validator(holder_id)

        # Use simple string variant if possible
        if not self._holder or isinstance(self._holder, str):
            self._holder = holder_id
        else:
            self._holder["id"] = holder_id

    @property
    def holder(self):
        """Getter for holder."""
        return self._holder

    @holder.setter
    def holder(self, holder: Union[str, dict]):
        """Setter for holder."""
        uri_validator = Uri()

        holder_id = holder if isinstance(holder, str) else holder.get("id")

        if not holder_id:
            raise ValidationError("Holder id is required")
        uri_validator(holder_id)

        self._holder = holder

    @property
    def verifiable_credential(self):
        """Getter for verifiable credential."""
        return self._verifiable_credential

    @verifiable_credential.setter
    def verifiable_credential(self, verifiable_credential: List[dict]):
        """Setter for verifiable credential."""

        self._verifiable_credential = verifiable_credential

    @property
    def proof(self):
        """Getter for proof."""
        return self._proof

    @proof.setter
    def proof(self, proof: LDProof):
        """Setter for proof."""
        self._proof = proof

    def __eq__(self, o: object) -> bool:
        """Check equality."""
        if isinstance(o, VerifiablePresentation):
            return (
                self.context == o.context
                and self.id == o.id
                and self.type == o.type
                and self.holder == o.holder
                and self.verifiable_credentials == o.verifiable_credentials
                and self.proof == o.proof
                and self.extra == o.extra
            )

        return False


class PresentationSchema(BaseModelSchema):
    """Linked data presentation schema.

    Based on https://www.w3.org/TR/vc-data-model

    """

    class Meta:
        """Accept parameter overload."""

        unknown = INCLUDE
        model_class = VerifiablePresentation

    context = fields.List(
        UriOrDictField(required=True),
        data_key="@context",
        required=True,
        validate=CREDENTIAL_CONTEXT_VALIDATE,
        metadata={
            "description": "The JSON-LD context of the presentation",
            "example": [CREDENTIALS_CONTEXT_V1_URL],
        },
    )

    id = fields.Str(
        required=False,
        validate=Uri(),
        metadata={
            "description": "The ID of the presentation",
            "example": "http://example.edu/presentations/1872",
        },
    )

    type = fields.List(
        fields.Str(required=True),
        required=True,
        validate=PRESENTATION_TYPE_VALIDATE,
        metadata={
            "description": "The JSON-LD type of the presentation",
            "example": PRESENTATION_TYPE_EXAMPLE,
        },
    )

    holder = StrOrDictField(
        required=False,
        metadata={
            "description": (
                "The JSON-LD Verifiable Credential Holder. Either string of object with"
                " id field."
            ),
            "example": DIDKey.EXAMPLE,
        },
    )

    # TODO how to validate VCs in list
    verifiable_credential = fields.List(
        fields.Dict(required=True),
        required=False,
        data_key="verifiableCredential",
        metadata={},
    )

    proof = fields.Nested(
        LinkedDataProofSchema(),
        required=False,
        metadata={
            "description": "The proof of the presentation",
            "example": {
                "type": "Ed25519Signature2018",
                "verificationMethod": (
                    "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38Ee"
                    "fXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                ),
                "created": "2019-12-11T03:50:55",
                "proofPurpose": "assertionMethod",
                "jws": (
                    "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0JiNjQiXX0..lKJU0Df_k"
                    "eblRKhZAS9Qq6zybm-HqUXNVZ8vgEPNTAjQKBhQDxvXNo7nvtUBb_Eq1Ch6YBKY5qBQ"
                ),
            },
        },
    )

    @post_dump(pass_original=True)
    def add_unknown_properties(self, data: dict, original, **kwargs):
        """Add back unknown properties before outputting."""

        data.update(original.extra)

        return data


class VerifiablePresentationSchema(PresentationSchema):
    """Single Verifiable Presentation Schema."""

    proof = fields.Nested(
        LinkedDataProofSchema(),
        required=True,
        metadata={
            "description": "The proof of the presentation",
            "example": {
                "type": "Ed25519Signature2018",
                "verificationMethod": (
                    "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38Ee"
                    "fXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                ),
                "created": "2019-12-11T03:50:55",
                "proofPurpose": "assertionMethod",
                "jws": (
                    "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0JiNjQiXX0..lKJU0Df_k"
                    "eblRKhZAS9Qq6zybm-HqUXNVZ8vgEPNTAjQKBhQDxvXNo7nvtUBb_Eq1Ch6YBKY5qBQ"
                ),
            },
        },
    )



"""Class for paginated query parameters."""

from typing import Tuple

from aiohttp.web import BaseRequest
from marshmallow import fields
from marshmallow.validate import OneOf, Range

from ...messaging.models.openapi import OpenAPISchema
from ...storage.base import DEFAULT_PAGE_SIZE, MAXIMUM_PAGE_SIZE


class PaginatedQuerySchema(OpenAPISchema):
    """Parameters for paginated queries."""

    limit = fields.Int(
        required=False,
        load_default=DEFAULT_PAGE_SIZE,
        validate=Range(min=1, max=MAXIMUM_PAGE_SIZE),
        metadata={"description": "Number of results to return", "example": 50},
    )
    offset = fields.Int(
        required=False,
        load_default=0,
        validate=Range(min=0),
        metadata={"description": "Offset for pagination", "example": 0},
    )
    order_by = fields.Str(
        required=False,
        load_default="id",
        validate=OneOf(["id"]),  # only one possible column supported in askar
        metadata={
            "description": (
                'The column to order results by. Only "id" is currently supported.'
            )
        },
        error_messages={"validator_failed": '`order_by` only supports column "id"'},
    )
    descending = fields.Bool(
        required=False,
        load_default=False,
        truthy={"true", "1", "yes"},
        falsy={"false", "0", "no"},
        metadata={"description": "Order results in descending order if true"},
        error_messages={"invalid": "Not a valid boolean."},
    )


def get_paginated_query_params(request: BaseRequest) -> Tuple[int, int, str, bool]:
    """Read the limit, offset, order_by, and descending query parameters from a request.

    Args:
        request: aiohttp request object.

    Returns:
        A tuple containing:
        - limit (int): The number of results to return, defaulting to DEFAULT_PAGE_SIZE.
        - offset (int): The offset for pagination, defaulting to 0.
        - order_by (str): The field by which to order results, defaulting to "id".
        - descending (bool): Order results in descending order; defaults to False.
    """

    limit = int(request.query.get("limit", DEFAULT_PAGE_SIZE))
    offset = int(request.query.get("offset", 0))
    order_by = request.query.get("order_by", "id")

    # Convert the 'descending' parameter to a boolean
    descending_str = request.query.get("descending", "False").lower()
    descending = descending_str in {"true", "1", "yes"}

    return limit, offset, order_by, descending



from unittest import TestCase

from ..please_ack_decorator import PleaseAckDecorator

MESSAGE_ID = "abc123"
ON = ("RECEIPT", "OUTCOME")


class TestPleaseAckDecorator(TestCase):
    def test_init_serde(self):
        decorator = PleaseAckDecorator()
        assert type(decorator) is PleaseAckDecorator
        assert decorator.message_id is None
        assert decorator.on is None
        dumped = decorator.serialize()
        assert dumped == {}
        loaded = PleaseAckDecorator.deserialize(dumped)
        assert type(loaded) is PleaseAckDecorator
        assert loaded.message_id is None
        assert loaded.on is None

        decorator = PleaseAckDecorator(message_id=MESSAGE_ID)
        assert type(decorator) is PleaseAckDecorator
        assert decorator.message_id == MESSAGE_ID
        assert decorator.on is None
        dumped = decorator.serialize()
        assert dumped == {"message_id": MESSAGE_ID}
        loaded = PleaseAckDecorator.deserialize(dumped)
        assert type(loaded) is PleaseAckDecorator
        assert loaded.message_id == MESSAGE_ID
        assert loaded.on is None

        decorator = PleaseAckDecorator(on=ON)
        assert type(decorator) is PleaseAckDecorator
        assert decorator.message_id is None
        assert decorator.on == list(ON)
        dumped = decorator.serialize()
        assert dumped == {
            "on": list(ON),
        }
        loaded = PleaseAckDecorator.deserialize(dumped)
        assert type(loaded) is PleaseAckDecorator
        assert loaded.message_id is None
        assert loaded.on == list(ON)

        decorator = PleaseAckDecorator(message_id=MESSAGE_ID, on=ON)
        assert type(decorator) is PleaseAckDecorator
        assert decorator.message_id == MESSAGE_ID
        assert decorator.on == list(ON)
        dumped = decorator.serialize()
        assert dumped == {
            "message_id": MESSAGE_ID,
            "on": list(ON),
        }
        loaded = PleaseAckDecorator.deserialize(dumped)
        assert type(loaded) is PleaseAckDecorator
        assert loaded.message_id == MESSAGE_ID
        assert loaded.on == list(ON)



"""A message decorator for services.

A service decorator adds routing information to a message so agent can respond without
needing to perform a handshake.
"""

from typing import List, Optional

from marshmallow import EXCLUDE, fields

from ..models.base import BaseModel, BaseModelSchema
from ..valid import (
    RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
    RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
)


class ServiceDecorator(BaseModel):
    """Class representing service decorator."""

    class Meta:
        """ServiceDecorator metadata."""

        schema_class = "ServiceDecoratorSchema"

    def __init__(
        self,
        *,
        endpoint: str,
        recipient_keys: List[str],
        routing_keys: Optional[List[str]] = None,
    ):
        """Initialize a ServiceDecorator instance.

        Args:
            endpoint: Endpoint which this agent can be reached at
            recipient_keys: List of recipient keys
            routing_keys: List of routing keys

        """
        super().__init__()
        self._endpoint = endpoint
        self._recipient_keys = recipient_keys
        self._routing_keys = routing_keys

    @property
    def endpoint(self):
        """Accessor for service endpoint.

        Returns:
            This service's `serviceEndpoint`

        """
        return self._endpoint

    @property
    def recipient_keys(self):
        """Accessor for recipient keys.

        Returns:
            This service's `recipientKeys`

        """
        return self._recipient_keys

    @property
    def routing_keys(self):
        """Accessor for routing keys.

        Returns:
            This service's `routingKeys`

        """
        return self._routing_keys


class ServiceDecoratorSchema(BaseModelSchema):
    """Thread decorator schema used in serialization/deserialization."""

    class Meta:
        """ServiceDecoratorSchema metadata."""

        model_class = ServiceDecorator
        unknown = EXCLUDE

    recipient_keys = fields.List(
        fields.Str(
            validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
            metadata={
                "description": "Recipient public key",
                "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
            },
        ),
        data_key="recipientKeys",
        required=True,
        metadata={"description": "List of recipient keys"},
    )
    endpoint = fields.Str(
        data_key="serviceEndpoint",
        required=True,
        metadata={
            "description": "Service endpoint at which to reach this agent",
            "example": "http://192.168.56.101:8020",
        },
    )
    routing_keys = fields.List(
        fields.Str(
            validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
            metadata={
                "description": "Routing key",
                "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
            },
        ),
        data_key="routingKeys",
        required=False,
        metadata={"description": "List of routing keys"},
    )



"""Classes to manage discover features."""

import asyncio
import logging
from typing import Optional, Sequence, Tuple

from ....core.error import BaseError
from ....core.goal_code_registry import GoalCodeRegistry
from ....core.profile import Profile
from ....core.protocol_registry import ProtocolRegistry
from ....messaging.responder import BaseResponder
from ....storage.error import StorageNotFoundError
from .messages.disclosures import Disclosures
from .messages.queries import Queries, QueryItem
from .models.discovery_record import V20DiscoveryExchangeRecord


class V20DiscoveryMgrError(BaseError):
    """Discover feature v2_0 error."""


class V20DiscoveryMgr:
    """Class for discover feature v1_0 under RFC 31."""

    def __init__(self, profile: Profile):
        """Initialize a V20DiscoveryMgr.

        Args:
            profile: The profile for this manager
        """
        self._profile = profile
        self._logger = logging.getLogger(__name__)

    @property
    def profile(self) -> Profile:
        """Accessor for the current Profile.

        Returns:
            The Profile for this manager

        """
        return self._profile

    async def receive_disclose(
        self, disclose_msg: Disclosures, connection_id: Optional[str] = None
    ) -> V20DiscoveryExchangeRecord:
        """Receive Disclose message and return updated V20DiscoveryExchangeRecord."""
        if disclose_msg._thread:
            thread_id = disclose_msg._thread.thid
            try:
                async with self._profile.session() as session:
                    discover_exch_rec = await V20DiscoveryExchangeRecord.retrieve_by_id(
                        session=session, record_id=thread_id
                    )
            except StorageNotFoundError:
                discover_exch_rec = await self.lookup_exchange_rec_by_connection(
                    connection_id
                )
                if not discover_exch_rec:
                    discover_exch_rec = V20DiscoveryExchangeRecord()
        else:
            discover_exch_rec = await self.lookup_exchange_rec_by_connection(
                connection_id
            )
            if not discover_exch_rec:
                discover_exch_rec = V20DiscoveryExchangeRecord()
        async with self._profile.session() as session:
            discover_exch_rec.disclosures = disclose_msg
            discover_exch_rec.connection_id = connection_id
            await discover_exch_rec.save(session)
        return discover_exch_rec

    async def lookup_exchange_rec_by_connection(
        self, connection_id: str
    ) -> Optional[V20DiscoveryExchangeRecord]:
        """Retrieve V20DiscoveryExchangeRecord by connection_id."""
        async with self._profile.session() as session:
            if await V20DiscoveryExchangeRecord.exists_for_connection_id(
                session=session, connection_id=connection_id
            ):
                return await V20DiscoveryExchangeRecord.retrieve_by_connection_id(
                    session=session, connection_id=connection_id
                )
            else:
                return None

    async def proactive_disclose_features(self, connection_id: str):
        """Proactively disclose features on active connection setup."""
        queries_msg = Queries(
            queries=[
                QueryItem(feature_type="protocol", match="*"),
                QueryItem(feature_type="goal-code", match="*"),
            ]
        )
        disclosures = await self.receive_query(queries_msg=queries_msg)
        responder = self.profile.inject_or(BaseResponder)
        if responder:
            await responder.send(disclosures, connection_id=connection_id)
        else:
            self._logger.exception(
                "Unable to send discover-features v2 disclosures message"
                ": BaseResponder unavailable"
            )

    async def return_to_publish_features(
        self,
    ) -> Tuple[Optional[Sequence[str]], Optional[Sequence[str]]]:
        """Return to_publish features filter, if specified."""
        to_publish_protocols = None
        to_publish_goal_codes = None
        async with self._profile.session() as session:
            if (
                session.settings.get("disclose_protocol_list")
                and len(session.settings.get("disclose_protocol_list")) > 0
            ):
                to_publish_protocols = session.settings.get("disclose_protocol_list")
            if (
                session.settings.get("disclose_goal_code_list")
                and len(session.settings.get("disclose_goal_code_list")) > 0
            ):
                to_publish_goal_codes = session.settings.get("disclose_goal_code_list")
        return (to_publish_protocols, to_publish_goal_codes)

    async def execute_protocol_query(self, match: str):
        """Execute protocol match query."""
        protocol_registry = self._profile.context.inject_or(ProtocolRegistry)
        protocols = protocol_registry.protocols_matching_query(match)
        results = await protocol_registry.prepare_disclosed(
            self._profile.context, protocols
        )
        return results

    async def execute_goal_code_query(self, match: str):
        """Execute goal code match query."""
        goal_code_registry = self._profile.context.inject_or(GoalCodeRegistry)
        results = goal_code_registry.goal_codes_matching_query(match)
        return results

    async def receive_query(self, queries_msg: Queries) -> Disclosures:
        """Process query and return the corresponding disclose message."""
        disclosures = Disclosures(disclosures=[])
        published_results = []
        (
            to_publish_protocols,
            to_publish_goal_codes,
        ) = await self.return_to_publish_features()
        for query_item in queries_msg.queries:
            assert isinstance(query_item, QueryItem)
            if query_item.feature_type == "protocol":
                results = await self.execute_protocol_query(query_item.match)
                for result in results:
                    to_publish_result = {"feature-type": "protocol"}
                    if "pid" in result:
                        if (
                            to_publish_protocols
                            and result.get("pid") not in to_publish_protocols
                        ):
                            continue
                        to_publish_result["id"] = result.get("pid")
                    else:
                        continue
                    if "roles" in result:
                        to_publish_result["roles"] = result.get("roles")
                    published_results.append(to_publish_result)
            elif query_item.feature_type == "goal-code":
                results = await self.execute_goal_code_query(query_item.match)
                for result in results:
                    to_publish_result = {"feature-type": "goal-code"}
                    if to_publish_goal_codes and result not in to_publish_goal_codes:
                        continue
                    to_publish_result["id"] = result
                    published_results.append(to_publish_result)
        disclosures.disclosures = published_results
        # Check if query message has a thid
        # If disclosing this agents feature
        if queries_msg._thread:
            disclosures.assign_thread_id(queries_msg._thread.thid)
        return disclosures

    async def check_if_disclosure_received(
        self, record_id: str
    ) -> V20DiscoveryExchangeRecord:
        """Check if disclosures has been received."""
        while True:
            async with self._profile.session() as session:
                ex_rec = await V20DiscoveryExchangeRecord.retrieve_by_id(
                    session=session, record_id=record_id
                )
            if ex_rec.disclosures:
                return ex_rec
            await asyncio.sleep(0.5)

    async def create_and_send_query(
        self,
        connection_id: Optional[str] = None,
        query_protocol: Optional[str] = None,
        query_goal_code: Optional[str] = None,
    ) -> V20DiscoveryExchangeRecord:
        """Create and send a Query message."""
        queries = []
        if not query_goal_code and not query_protocol:
            raise V20DiscoveryMgrError(
                "At least one protocol or goal-code feature-type query is required."
            )
        if query_protocol:
            queries.append(QueryItem(feature_type="protocol", match=query_protocol))
        if query_goal_code:
            queries.append(QueryItem(feature_type="goal-code", match=query_goal_code))
        queries_msg = Queries(queries=queries)
        if connection_id:
            async with self._profile.session() as session:
                # If existing record exists for a connection_id
                if await V20DiscoveryExchangeRecord.exists_for_connection_id(
                    session=session, connection_id=connection_id
                ):
                    discovery_ex_rec = (
                        await V20DiscoveryExchangeRecord.retrieve_by_connection_id(
                            session=session, connection_id=connection_id
                        )
                    )
                    discovery_ex_rec.disclosures = None
                    await discovery_ex_rec.save(session)
                else:
                    discovery_ex_rec = V20DiscoveryExchangeRecord()
                discovery_ex_rec.queries_msg = queries_msg
                discovery_ex_rec.connection_id = connection_id
                await discovery_ex_rec.save(session)
            queries_msg.assign_thread_id(discovery_ex_rec.discovery_exchange_id)
            responder = self.profile.inject_or(BaseResponder)
            if responder:
                await responder.send(queries_msg, connection_id=connection_id)
            else:
                self._logger.exception(
                    "Unable to send discover-features v2 query message"
                    ": BaseResponder unavailable"
                )
            try:
                return await asyncio.wait_for(
                    self.check_if_disclosure_received(
                        record_id=discovery_ex_rec.discovery_exchange_id,
                    ),
                    5,
                )
            except asyncio.TimeoutError:
                return discovery_ex_rec
        else:
            # Disclose this agent's features and/or goal codes
            discovery_ex_rec = V20DiscoveryExchangeRecord()
            discovery_ex_rec.queries_msg = queries_msg
            disclosures = await self.receive_query(queries_msg=queries_msg)
            discovery_ex_rec.disclosures = disclosures
            return discovery_ex_rec



import json
from typing import Optional
from unittest import IsolatedAsyncioTestCase

import pytest
from marshmallow import EXCLUDE

from ...cache.base import BaseCache
from ...cache.in_memory import InMemoryCache
from ...config.injection_context import InjectionContext
from ...core.event_bus import EventBus
from ...core.protocol_registry import ProtocolRegistry
from ...messaging.agent_message import AgentMessage, AgentMessageSchema
from ...messaging.request_context import RequestContext
from ...protocols.coordinate_mediation.v1_0.route_manager import RouteManager
from ...protocols.didcomm_prefix import DIDCommPrefix
from ...protocols.issue_credential.v2_0.message_types import CRED_20_PROBLEM_REPORT
from ...protocols.issue_credential.v2_0.messages.cred_problem_report import (
    V20CredProblemReport,
)
from ...protocols.problem_report.v1_0.message import ProblemReport
from ...tests import mock
from ...transport.inbound.message import InboundMessage
from ...transport.inbound.receipt import MessageReceipt
from ...transport.outbound.message import OutboundMessage
from ...utils.stats import Collector
from ...utils.testing import create_test_profile
from .. import dispatcher as test_module


def make_inbound(payload) -> InboundMessage:
    return InboundMessage(payload, MessageReceipt(thread_id="dummy-thread"))


class Receiver:
    def __init__(self):
        self.messages = []

    async def send(
        self,
        context: InjectionContext,
        message: OutboundMessage,
        inbound: Optional[InboundMessage] = None,
    ):
        self.messages.append((context, message, inbound))


class StubAgentMessage(AgentMessage):
    class Meta:
        handler_class = "StubAgentMessageHandler"
        schema_class = "StubAgentMessageSchema"
        message_type = "doc/proto-name/1.1/message-type"


class StubAgentMessageSchema(AgentMessageSchema):
    class Meta:
        model_class = StubAgentMessage
        unknown = EXCLUDE


class StubAgentMessageHandler:
    async def handle(self, context, responder):
        pass


class StubV1_2AgentMessage(AgentMessage):
    class Meta:
        handler_class = "StubV1_2AgentMessageHandler"
        schema_class = "StubV1_2AgentMessageSchema"
        message_type = "doc/proto-name/1.2/message-type"


class StubV1_2AgentMessageSchema(AgentMessageSchema):
    class Meta:
        model_class = StubV1_2AgentMessage
        unknonw = EXCLUDE


class StubV1_2AgentMessageHandler:
    async def handle(self, context, responder):
        pass


class TestDispatcher(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.profile.context.injector.bind_instance(ProtocolRegistry, ProtocolRegistry())
        self.profile.context.injector.bind_instance(Collector, Collector())
        self.profile.context.injector.bind_instance(EventBus, EventBus())
        self.profile.context.injector.bind_instance(RouteManager, mock.MagicMock())

    async def test_dispatch(self):
        profile = self.profile
        registry = profile.inject(ProtocolRegistry)
        registry.register_message_types(
            {
                pfx.qualify(StubAgentMessage.Meta.message_type): StubAgentMessage
                for pfx in DIDCommPrefix
            }
        )
        dispatcher = test_module.Dispatcher(profile)
        await dispatcher.setup()
        rcv = Receiver()
        message = {
            "@type": DIDCommPrefix.qualify_current(StubAgentMessage.Meta.message_type)
        }

        with (
            mock.patch.object(
                StubAgentMessageHandler, "handle", autospec=True
            ) as handler_mock,
            mock.patch.object(
                test_module, "BaseConnectionManager", autospec=True
            ) as conn_mgr_mock,
        ):
            conn_mgr_mock.return_value = mock.MagicMock(
                find_inbound_connection=mock.CoroutineMock(
                    return_value=mock.MagicMock(connection_id="dummy")
                )
            )
            await dispatcher.queue_message(
                dispatcher.profile, make_inbound(message), rcv.send
            )
            await dispatcher.task_queue
            handler_mock.assert_awaited_once()
            assert isinstance(handler_mock.call_args[0][1].message, StubAgentMessage)
            assert isinstance(
                handler_mock.call_args[0][2], test_module.DispatcherResponder
            )

    async def test_dispatch_versioned_message(self):
        profile = self.profile
        registry = profile.inject(ProtocolRegistry)
        registry.register_message_types(
            {
                DIDCommPrefix.qualify_current(
                    StubAgentMessage.Meta.message_type
                ): StubAgentMessage
            },
            version_definition={
                "major_version": 1,
                "minimum_minor_version": 0,
                "current_minor_version": 1,
                "path": "v1_1",
            },
        )
        dispatcher = test_module.Dispatcher(profile)
        await dispatcher.setup()
        rcv = Receiver()
        message = {
            "@type": DIDCommPrefix.qualify_current(StubAgentMessage.Meta.message_type)
        }

        with (
            mock.patch.object(
                StubAgentMessageHandler, "handle", autospec=True
            ) as handler_mock,
            mock.patch.object(test_module, "BaseConnectionManager", autospec=True),
        ):
            await dispatcher.queue_message(
                dispatcher.profile, make_inbound(message), rcv.send
            )
            await dispatcher.task_queue
            handler_mock.assert_awaited_once()
            assert isinstance(handler_mock.call_args[0][1].message, StubAgentMessage)
            assert isinstance(
                handler_mock.call_args[0][2], test_module.DispatcherResponder
            )

    @pytest.mark.skip("This test is not completing")
    async def test_dispatch_versioned_message_no_message_class(self):
        registry = self.profile.inject(ProtocolRegistry)
        registry.register_message_types(
            {
                DIDCommPrefix.qualify_current(
                    StubAgentMessage.Meta.message_type
                ): StubAgentMessage
            },
            version_definition={
                "major_version": 1,
                "minimum_minor_version": 0,
                "current_minor_version": 1,
                "path": "v1_1",
            },
        )
        dispatcher = test_module.Dispatcher(self.profile)
        await dispatcher.setup()
        rcv = Receiver()
        message = {"@type": "doc/proto-name/1.1/no-such-message-type"}

        with mock.patch.object(StubAgentMessageHandler, "handle", autospec=True):
            await dispatcher.queue_message(
                dispatcher.profile, make_inbound(message), rcv.send
            )
            await dispatcher.task_queue
            assert rcv.messages and isinstance(rcv.messages[0][1], OutboundMessage)
            payload = json.loads(rcv.messages[0][1].payload)
            assert payload["@type"] == DIDCommPrefix.qualify_current(
                ProblemReport.Meta.message_type
            )

    @pytest.mark.skip("This test is not completing")
    async def test_dispatch_versioned_message_message_class_deserialize_x(self):
        profile = self.profile
        registry = profile.inject(ProtocolRegistry)
        registry.register_message_types(
            {
                DIDCommPrefix.qualify_current(
                    StubAgentMessage.Meta.message_type
                ): StubAgentMessage
            },
            version_definition={
                "major_version": 1,
                "minimum_minor_version": 0,
                "current_minor_version": 1,
                "path": "v1_1",
            },
        )
        dispatcher = test_module.Dispatcher(profile)
        await dispatcher.setup()
        rcv = Receiver()
        message = {"@type": "doc/proto-name/1.1/no-such-message-type"}

        with (
            mock.patch.object(StubAgentMessageHandler, "handle", autospec=True),
            mock.patch.object(
                registry, "resolve_message_class", mock.MagicMock()
            ) as mock_resolve,
        ):
            mock_resolve.return_value = mock.MagicMock(
                deserialize=mock.MagicMock(side_effect=test_module.BaseModelError())
            )
            await dispatcher.queue_message(
                dispatcher.profile, make_inbound(message), rcv.send
            )
            await dispatcher.task_queue
            assert rcv.messages and isinstance(rcv.messages[0][1], OutboundMessage)
            payload = json.loads(rcv.messages[0][1].payload)
            assert payload["@type"] == DIDCommPrefix.qualify_current(
                ProblemReport.Meta.message_type
            )

    async def test_dispatch_versioned_message_handle_greater_succeeds(self):
        profile = self.profile
        registry = profile.inject(ProtocolRegistry)
        registry.register_message_types(
            {
                DIDCommPrefix.qualify_current(
                    StubAgentMessage.Meta.message_type
                ): StubAgentMessage
            },
            version_definition={
                "major_version": 1,
                "minimum_minor_version": 0,
                "current_minor_version": 1,
                "path": "v1_1",
            },
        )
        dispatcher = test_module.Dispatcher(profile)
        await dispatcher.setup()
        rcv = Receiver()
        message = {
            "@type": DIDCommPrefix.qualify_current(StubV1_2AgentMessage.Meta.message_type)
        }

        with (
            mock.patch.object(
                StubAgentMessageHandler, "handle", autospec=True
            ) as handler_mock,
            mock.patch.object(test_module, "BaseConnectionManager", autospec=True),
        ):
            await dispatcher.queue_message(
                dispatcher.profile, make_inbound(message), rcv.send
            )
            await dispatcher.task_queue
            handler_mock.assert_awaited_once()
            assert isinstance(handler_mock.call_args[0][1].message, StubAgentMessage)
            assert isinstance(
                handler_mock.call_args[0][2], test_module.DispatcherResponder
            )

    @pytest.mark.skip("This test is not completing")
    async def test_dispatch_versioned_message_fail(self):
        profile = self.profile
        registry = profile.inject(ProtocolRegistry)
        registry.register_message_types(
            {
                DIDCommPrefix.qualify_current(
                    StubV1_2AgentMessage.Meta.message_type
                ): StubV1_2AgentMessage
            },
            version_definition={
                "major_version": 1,
                "minimum_minor_version": 2,
                "current_minor_version": 2,
                "path": "v1_2",
            },
        )
        dispatcher = test_module.Dispatcher(profile)
        await dispatcher.setup()
        rcv = Receiver()
        message = {
            "@type": DIDCommPrefix.qualify_current(StubAgentMessage.Meta.message_type)
        }

        with mock.patch.object(StubAgentMessageHandler, "handle", autospec=True):
            await dispatcher.queue_message(
                dispatcher.profile, make_inbound(message), rcv.send
            )
            await dispatcher.task_queue
            assert rcv.messages and isinstance(rcv.messages[0][1], OutboundMessage)
            payload = json.loads(rcv.messages[0][1].payload)
            assert payload["@type"] == DIDCommPrefix.qualify_current(
                ProblemReport.Meta.message_type
            )

    @pytest.mark.skip("This test is not completing")
    async def test_bad_message_dispatch_parse_x(self):
        dispatcher = test_module.Dispatcher(self.profile)
        await dispatcher.setup()
        rcv = Receiver()
        bad_messages = ["not even a dict", {"bad": "message"}]
        for bad in bad_messages:
            await dispatcher.queue_message(
                dispatcher.profile, make_inbound(bad), rcv.send
            )
            await dispatcher.task_queue
            assert rcv.messages and isinstance(rcv.messages[0][1], OutboundMessage)
            payload = json.loads(rcv.messages[0][1].payload)
            assert payload["@type"] == DIDCommPrefix.qualify_current(
                ProblemReport.Meta.message_type
            )
            rcv.messages.clear()

    async def test_bad_message_dispatch_problem_report_x(self):
        profile = self.profile
        registry = profile.inject(ProtocolRegistry)
        registry.register_message_types(
            {
                pfx.qualify(CRED_20_PROBLEM_REPORT): V20CredProblemReport
                for pfx in DIDCommPrefix
            }
        )
        dispatcher = test_module.Dispatcher(profile)
        await dispatcher.setup()
        rcv = Receiver()
        bad_message = {
            "@type": DIDCommPrefix.qualify_current(CRED_20_PROBLEM_REPORT),
            "description": "should be a dict",
        }
        await dispatcher.queue_message(
            dispatcher.profile, make_inbound(bad_message), rcv.send
        )
        await dispatcher.task_queue
        assert not rcv.messages

    async def test_dispatch_log(self):
        profile = self.profile
        registry = profile.inject(ProtocolRegistry)
        registry.register_message_types(
            {
                DIDCommPrefix.qualify_current(
                    StubAgentMessage.Meta.message_type
                ): StubAgentMessage
            },
        )

        dispatcher = test_module.Dispatcher(profile)
        await dispatcher.setup()

        exc = KeyError("sample exception")
        mock_task = mock.MagicMock(
            exc_info=(type(exc), exc, exc.__traceback__),
            ident="abc",
            timing={
                "queued": 1234567890,
                "unqueued": 1234567899,
                "started": 1234567901,
                "ended": 1234567999,
            },
        )
        dispatcher.log_task(mock_task)

    async def test_create_send_outbound(self):
        profile = self.profile
        context = RequestContext(
            profile,
            settings={"timing.enabled": True},
        )
        registry = profile.inject(ProtocolRegistry)
        registry.register_message_types(
            {
                pfx.qualify(StubAgentMessage.Meta.message_type): StubAgentMessage
                for pfx in DIDCommPrefix
            }
        )
        message = StubAgentMessage()
        responder = test_module.DispatcherResponder(context, message, None)
        outbound_message = await responder.create_outbound(
            json.dumps(message.serialize())
        )
        with (
            mock.patch.object(responder, "_send", mock.CoroutineMock()),
            mock.patch.object(
                test_module.BaseResponder,
                "conn_rec_active_state_check",
                mock.CoroutineMock(return_value=True),
            ),
        ):
            await responder.send_outbound(outbound_message)

    async def test_create_send_outbound_with_msg_attrs(self):
        profile = self.profile
        context = RequestContext(
            profile,
            settings={"timing.enabled": True},
        )
        registry = profile.inject(ProtocolRegistry)
        registry.register_message_types(
            {
                pfx.qualify(StubAgentMessage.Meta.message_type): StubAgentMessage
                for pfx in DIDCommPrefix
            }
        )
        message = StubAgentMessage()
        responder = test_module.DispatcherResponder(context, message, None)
        outbound_message = await responder.create_outbound(message)
        with (
            mock.patch.object(responder, "_send", mock.CoroutineMock()),
            mock.patch.object(
                test_module.BaseResponder,
                "conn_rec_active_state_check",
                mock.CoroutineMock(return_value=True),
            ),
        ):
            await responder.send_outbound(
                message=outbound_message,
                message_type=message._message_type,
                message_id=message._id,
            )

    async def test_create_send_outbound_with_msg_attrs_x(self):
        profile = self.profile
        context = RequestContext(
            profile,
            settings={"timing.enabled": True},
        )
        registry = profile.inject(ProtocolRegistry)
        registry.register_message_types(
            {
                pfx.qualify(StubAgentMessage.Meta.message_type): StubAgentMessage
                for pfx in DIDCommPrefix
            }
        )
        message = StubAgentMessage()
        responder = test_module.DispatcherResponder(context, message, None)
        outbound_message = await responder.create_outbound(message)
        outbound_message.connection_id = "123"
        with mock.patch.object(
            test_module.BaseResponder,
            "conn_rec_active_state_check",
            mock.CoroutineMock(return_value=False),
        ):
            with self.assertRaises(RuntimeError):
                await responder.send_outbound(
                    message=outbound_message,
                    message_type=message._message_type,
                    message_id=message._id,
                )

    async def test_create_send_webhook(self):
        profile = self.profile
        context = RequestContext(profile)
        message = StubAgentMessage()
        responder = test_module.DispatcherResponder(context, message, None)
        with pytest.deprecated_call():
            await responder.send_webhook("topic", {"pay": "load"})

    async def test_conn_rec_active_state_check_a(self):
        profile = self.profile
        profile.context.injector.bind_instance(BaseCache, InMemoryCache())
        context = RequestContext(profile)
        message = StubAgentMessage()
        responder = test_module.DispatcherResponder(context, message, None)
        with mock.patch.object(
            test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_ret_by_id:
            conn_rec = test_module.ConnRecord()
            conn_rec.state = test_module.ConnRecord.State.COMPLETED
            mock_conn_ret_by_id.return_value = conn_rec
            check_flag = await responder.conn_rec_active_state_check(
                profile,
                "conn-id",
            )
            assert check_flag
            check_flag = await responder.conn_rec_active_state_check(
                profile,
                "conn-id",
            )
            assert check_flag

    async def test_conn_rec_active_state_check_b(self):
        profile = self.profile
        profile.context.injector.bind_instance(BaseCache, InMemoryCache())
        profile.context.injector.bind_instance(
            EventBus, mock.MagicMock(notify=mock.CoroutineMock())
        )
        context = RequestContext(profile)
        message = StubAgentMessage()
        responder = test_module.DispatcherResponder(context, message, None)
        with mock.patch.object(
            test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_ret_by_id:
            conn_rec_a = test_module.ConnRecord()
            conn_rec_a.state = test_module.ConnRecord.State.REQUEST
            conn_rec_b = test_module.ConnRecord()
            conn_rec_b.state = test_module.ConnRecord.State.COMPLETED
            mock_conn_ret_by_id.side_effect = [conn_rec_a, conn_rec_b]
            check_flag = await responder.conn_rec_active_state_check(
                profile,
                "conn-id",
            )
            assert check_flag

    async def test_create_enc_outbound(self):
        profile = self.profile
        context = RequestContext(profile)
        message = StubAgentMessage()
        responder = test_module.DispatcherResponder(context, message, None)
        with mock.patch.object(
            responder, "send_outbound", mock.CoroutineMock()
        ) as mock_send_outbound:
            await responder.send(message)
            mock_send_outbound.assert_called_once()
        msg_json = json.dumps(StubAgentMessage().serialize())
        message = msg_json.encode("utf-8")
        with mock.patch.object(
            responder, "send_outbound", mock.CoroutineMock()
        ) as mock_send_outbound:
            await responder.send(message)

        message = StubAgentMessage()
        with mock.patch.object(
            responder, "send_outbound", mock.CoroutineMock()
        ) as mock_send_outbound:
            await responder.send_reply(message)
            mock_send_outbound.assert_called_once()

        message = json.dumps(StubAgentMessage().serialize())
        with mock.patch.object(
            responder, "send_outbound", mock.CoroutineMock()
        ) as mock_send_outbound:
            await responder.send_reply(message)

    async def test_expired_context_x(self):
        def _smaller_scope():
            profile = self.profile
            context = RequestContext(profile)
            message = b"abc123xyz7890000"
            return test_module.DispatcherResponder(context, message, None)

        responder = _smaller_scope()
        with self.assertRaises(RuntimeError):
            await responder.create_outbound(b"test")

        with self.assertRaises(RuntimeError):
            await responder.send_outbound(None)

        with pytest.deprecated_call():
            with self.assertRaises(RuntimeError):
                await responder.send_webhook("test", {})



"""DID Rotate manager.

Manages and tracks the state of the DID Rotate protocol.
"""

from ....connections.base_manager import BaseConnectionManager, BaseConnectionManagerError
from ....connections.models.conn_record import ConnRecord
from ....core.profile import Profile
from ....messaging.responder import BaseResponder
from ....resolver.base import DIDMethodNotSupported, DIDNotFound
from ....resolver.did_resolver import DIDResolver
from .messages import Hangup, Rotate, RotateAck, RotateProblemReport
from .models import RotateRecord


class DIDRotateManagerError(Exception):
    """Raised when an error occurs during a DID Rotate protocol flow."""


class ReportableDIDRotateError(DIDRotateManagerError):
    """Base class for reportable errors."""

    def __init__(self, message: RotateProblemReport):
        """Initialize the ReportableDIDRotateError."""
        self.message = message


class UnresolvableDIDError(ReportableDIDRotateError):
    """Raised when a DID cannot be resolved."""


class UnsupportedDIDMethodError(ReportableDIDRotateError):
    """Raised when a DID method is not supported."""


class UnresolvableDIDCommServicesError(ReportableDIDRotateError):
    """Raised when DIDComm services cannot be resolved."""


class UnrecordableKeysError(ReportableDIDRotateError):
    """Raised when keys cannot be recorded for a resolvable DID."""


class DIDRotateManager:
    """DID Rotate Manager.

    Manages and tracks the state of the DID Rotate protocol.

    This mechanism allows a party in a relationship to change the DID they use
    to identify themselves in that relationship. This may be used to switch DID
    methods, but also to switch to a new DID within the same DID method. For
    non-updatable DID methods, this allows updating DID Doc attributes such as
    service endpoints. Inspired by (but different from) the DID rotation
    feature of the DIDComm Messaging (DIDComm v2) spec.

    DID Rotation is a pre-rotate operation. We send notification of rotation
    to the observing party before we rotate the DID. This allows the observing
    party to update their DID for the rotating party and notify the rotating
    party of any issues with the received DID.

    DID Rotation has two roles: the rotating party and the observing party.

    This manager is responsible for both of the possible roles in the protocol.
    """

    def __init__(self, profile: Profile):
        """Initialize DID Rotate Manager."""
        self.profile = profile

    async def hangup(self, conn: ConnRecord) -> Hangup:
        """Hangup the connection.

        Args:
            conn (ConnRecord): The connection to hangup.
        """

        hangup = Hangup()

        responder = self.profile.inject(BaseResponder)
        await responder.send(hangup, connection_id=conn.connection_id)

        async with self.profile.session() as session:
            await conn.delete_record(session)

        return hangup

    async def rotate_my_did(self, conn: ConnRecord, new_did: str) -> Rotate:
        """Rotate my DID.

        Args:
            conn (ConnRecord): The connection to rotate the DID for.
            new_did (str): The new DID to use for the connection.
        """

        record = RotateRecord(
            role=RotateRecord.ROLE_ROTATING,
            state=RotateRecord.STATE_ROTATE_SENT,
            connection_id=conn.connection_id,
            new_did=new_did,
        )
        rotate = Rotate(to_did=new_did)
        record.thread_id = rotate._message_id

        responder = self.profile.inject(BaseResponder)
        await responder.send(rotate, connection_id=conn.connection_id)

        async with self.profile.session() as session:
            await record.save(session, reason="Sent rotate message")

        return rotate

    async def receive_rotate(self, conn: ConnRecord, rotate: Rotate) -> RotateRecord:
        """Receive rotate message.

        Args:
            conn (ConnRecord): The connection to rotate the DID for.
            rotate (Rotate): The received rotate message.
        """
        record = RotateRecord(
            role=RotateRecord.ROLE_OBSERVING,
            state=RotateRecord.STATE_ROTATE_RECEIVED,
            connection_id=conn.connection_id,
            new_did=rotate.to_did,
            thread_id=rotate._message_id,
        )

        try:
            await self.ensure_supported_did(rotate.to_did)
        except ReportableDIDRotateError as err:
            responder = self.profile.inject(BaseResponder)
            err.message.assign_thread_from(rotate)
            await responder.send(err.message, connection_id=conn.connection_id)

        async with self.profile.session() as session:
            await record.save(session, reason="Received rotate message")

        return record

    async def commit_rotate(self, conn: ConnRecord, record: RotateRecord):
        """Commit rotate.

        Args:
            conn (ConnRecord): The connection to rotate the DID for.
            record (RotateRecord): The rotate record.
        """
        record.state = RotateRecord.STATE_ACK_SENT
        if not record.new_did:
            raise ValueError("No new DID stored in record")

        conn_mgr = BaseConnectionManager(self.profile)
        try:
            await conn_mgr.record_keys_for_resolvable_did(record.new_did)
        except BaseConnectionManagerError:
            raise UnrecordableKeysError(
                RotateProblemReport.unrecordable_keys(record.new_did)
            )

        conn.their_did = record.new_did

        ack = RotateAck()
        ack.assign_thread_id(thid=record.thread_id)

        responder = self.profile.inject(BaseResponder)
        await responder.send(ack, connection_id=conn.connection_id)

        async with self.profile.session() as session:
            # Don't emit a connection event for this change
            # Controllers should listen for the rotate event instead
            await conn.save(session, reason="Their DID rotated", event=False)
            await record.save(session, reason="Sent rotate ack")

        # TODO it would be better if the cache key included DIDs so we don't
        # have to manually clear it. This is a bigger change than a first pass
        # warrants though.
        await conn_mgr.clear_connection_targets_cache(conn.connection_id)

    async def receive_ack(self, conn: ConnRecord, ack: RotateAck):
        """Receive rotate ack message.

        Args:
            conn (ConnRecord): The connection to rotate the DID for.
            ack (RotateAck): The received rotate ack message.
        """
        async with self.profile.session() as session:
            record = await RotateRecord.retrieve_by_thread_id(session, ack._thread_id)

        record.state = RotateRecord.STATE_ACK_RECEIVED
        if not record.new_did:
            raise ValueError("No new DID stored in record")

        conn.my_did = record.new_did
        async with self.profile.session() as session:
            # Don't emit a connection event for this change
            # Controllers should listen for the rotate event instead
            await conn.save(session, reason="My DID rotated", event=False)
            # At this point the rotate is complete, so we can delete the record
            await record.delete_record(session)

        # TODO it would be better if the cache key included DIDs so we don't
        # have to manually clear it. This is a bigger change than a first pass
        # warrants though.
        conn_mgr = BaseConnectionManager(self.profile)
        await conn_mgr.clear_connection_targets_cache(conn.connection_id)

    async def receive_problem_report(self, problem_report: RotateProblemReport):
        """Receive problem report message.

        Args:
            conn (ConnRecord): The connection to rotate the DID for.
            problem_report (ProblemReport): The received problem report message.
        """
        async with self.profile.session() as session:
            record = await RotateRecord.retrieve_by_thread_id(
                session, problem_report._thread_id
            )

        record.state = RotateRecord.STATE_FAILED
        # Base ProblemReportSchema requires this value be present
        assert problem_report.description
        record.error = problem_report.description["code"]
        async with self.profile.session() as session:
            await record.save(session, reason="Received problem report")

    async def receive_hangup(self, conn: ConnRecord):
        """Receive hangup message.

        Args:
            conn (ConnRecord): The connection to rotate the DID for.
            hangup (Hangup): The received hangup message.
        """
        async with self.profile.session() as session:
            await conn.delete_record(session)

    async def ensure_supported_did(self, did: str):
        """Check if the DID is supported."""
        resolver = self.profile.inject(DIDResolver)
        conn_mgr = BaseConnectionManager(self.profile)
        try:
            await resolver.resolve(self.profile, did)
        except DIDMethodNotSupported:
            raise UnsupportedDIDMethodError(RotateProblemReport.unsupported_method(did))
        except DIDNotFound:
            raise UnresolvableDIDError(RotateProblemReport.unresolvable(did))

        try:
            await conn_mgr.resolve_didcomm_services(did)
        except BaseConnectionManagerError:
            raise UnresolvableDIDCommServicesError(
                RotateProblemReport.unresolvable_services(did)
            )



"""Commands module common setup."""

from importlib import import_module
from os import getenv
from typing import Sequence

PROG = getenv("ACAPY_COMMAND_NAME", "aca-py")


def available_commands():
    """Index available commands."""
    return [
        {"name": "help", "summary": "Print available commands"},
        {"name": "provision", "summary": "Provision an agent"},
        {"name": "start", "summary": "Start a new agent process"},
        {"name": "upgrade", "summary": "Start agent upgrade process"},
    ]


def load_command(command: str):
    """Load the module corresponding with a named command."""
    module = None
    module_path = None
    for cmd in available_commands():
        if cmd["name"] == command:
            module = cmd["name"]
            module_path = cmd.get("module")
            break
    if module and not module_path:
        module_path = f"{__package__}.{module}"
    if module_path:
        return import_module(module_path)


def run_command(command: str, argv: Sequence[str] = None):
    """Execute a named command with command line arguments."""
    module = load_command(command) or load_command("help")
    module.execute(argv)



"""Test VcLdpManager."""

from unittest import IsolatedAsyncioTestCase

import pytest

from acapy_agent.tests import mock
from acapy_agent.wallet.keys.manager import MultikeyManager

from ....did.did_key import DIDKey
from ....resolver.default.key import KeyDIDResolver
from ....resolver.did_resolver import DIDResolver
from ....storage.vc_holder.base import VCHolder
from ....utils.testing import create_test_profile
from ....wallet.base import BaseWallet
from ....wallet.default_verification_key_strategy import (
    BaseVerificationKeyStrategy,
    DefaultVerificationKeyStrategy,
)
from ....wallet.did_method import KEY, SOV, DIDMethod, DIDMethods
from ....wallet.key_type import BLS12381G2, ED25519, KeyTypes
from ...ld_proofs.constants import (
    SECURITY_CONTEXT_BBS_URL,
    SECURITY_CONTEXT_ED25519_2020_URL,
)
from ...ld_proofs.crypto.wallet_key_pair import WalletKeyPair
from ...ld_proofs.document_loader import DocumentLoader
from ...ld_proofs.purposes.authentication_proof_purpose import AuthenticationProofPurpose
from ...ld_proofs.purposes.credential_issuance_purpose import CredentialIssuancePurpose
from ...ld_proofs.suites.bbs_bls_signature_2020 import BbsBlsSignature2020
from ...ld_proofs.suites.bbs_bls_signature_proof_2020 import BbsBlsSignatureProof2020
from ...ld_proofs.suites.ecdsa_secp256r1_signature_2019 import EcdsaSecp256r1Signature2019
from ...ld_proofs.suites.ed25519_signature_2018 import Ed25519Signature2018
from ...ld_proofs.suites.ed25519_signature_2020 import Ed25519Signature2020
from ..manager import VcLdpManager, VcLdpManagerError
from ..models.credential import VerifiableCredential
from ..models.options import LDProofVCOptions

TEST_DID_SOV = "did:sov:LjgpST2rjsoxYegQDRm7EL"
TEST_DID_KEY = "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
TEST_DID_KEY_SEED = "testseed000000000000000000000001"
TEST_DID_KEY_VM = "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
TEST_DID_KEY_VERKEY = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
TEST_UUID = "urn:uuid:dc86e95c-dc85-4f91-b563-82657d095c44"
VC = {
    "credential": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1",
            {
                "ex": "https://example.org/test#",
                "test": "ex:test",
            },
        ],
        "type": ["VerifiableCredential", "UniversityDegreeCredential"],
        "credentialSubject": {"test": "key"},
        "issuanceDate": "2021-04-12",
        "issuer": TEST_DID_KEY,
    },
    "options": {
        "proofType": "Ed25519Signature2018",
        "created": "2019-12-11T03:50:55",
    },
}


class TestVcLdManager(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile()
        self.profile.context.injector.bind_instance(KeyTypes, KeyTypes())
        self.profile.context.injector.bind_instance(
            DIDResolver, DIDResolver([KeyDIDResolver()])
        )
        self.profile.context.injector.bind_instance(
            DefaultVerificationKeyStrategy, DefaultVerificationKeyStrategy()
        )
        self.profile.context.injector.bind_instance(
            BaseVerificationKeyStrategy,
            DefaultVerificationKeyStrategy(),
        )
        self.profile.context.injector.bind_instance(DIDMethods, DIDMethods())

        self.profile.context.injector.bind_instance(
            DocumentLoader, DocumentLoader(self.profile)
        )
        self.manager = VcLdpManager(self.profile)
        self.vc = VerifiableCredential.deserialize(VC["credential"])
        self.options = LDProofVCOptions.deserialize(VC["options"])

        async with self.profile.session() as session:
            await MultikeyManager(session=session).create(
                seed=TEST_DID_KEY_SEED, alg="ed25519"
            )

    methods: list[DIDMethod] = [SOV, KEY]

    async def test_get_did_info_for_did_sov(self):
        for method in self.methods:
            async with self.profile.session() as session:
                wallet = session.inject(BaseWallet)
                did = await wallet.create_local_did(
                    method=method,
                    key_type=ED25519,
                )
            did_info = await self.manager._did_info_for_did(did.did)
            assert did_info == did

    async def test_get_suite_for_document(self):
        with (
            mock.patch.object(
                self.manager,
                "_did_info_for_did",
                mock.CoroutineMock(),
            ) as mock_did_info,
        ):
            suite = await self.manager._get_suite_for_document(self.vc, self.options)

            assert suite.signature_type == self.options.proof_type
            assert isinstance(suite, Ed25519Signature2018)
            assert suite.verification_method == DIDKey.from_did(TEST_DID_KEY).key_id
            assert suite.proof == {"created": VC["options"]["created"]}
            assert isinstance(suite.key_pair, WalletKeyPair)
            assert suite.key_pair.key_type == ED25519
            assert suite.key_pair.public_key_base58 == TEST_DID_KEY_VERKEY

            mock_did_info.assert_awaited_once_with(self.vc.issuer)

    async def test_get_suite(self):
        proof = mock.MagicMock()
        did_info = mock.MagicMock()

        suite = await self.manager._get_suite(
            proof_type=BbsBlsSignature2020.signature_type,
            verification_method=TEST_DID_KEY_VM,
            proof=proof,
            did_info=did_info,
        )

        assert isinstance(suite, BbsBlsSignature2020)
        assert suite.verification_method == TEST_DID_KEY_VM
        assert suite.proof == proof
        assert isinstance(suite.key_pair, WalletKeyPair)
        assert suite.key_pair.key_type == BLS12381G2
        assert suite.key_pair.public_key_base58 == TEST_DID_KEY_VERKEY

        suite = await self.manager._get_suite(
            proof_type=Ed25519Signature2018.signature_type,
            verification_method=TEST_DID_KEY_VM,
            proof=proof,
            did_info=did_info,
        )

        assert isinstance(suite, Ed25519Signature2018)
        assert suite.verification_method == TEST_DID_KEY_VM
        assert suite.proof == proof
        assert isinstance(suite.key_pair, WalletKeyPair)
        assert suite.key_pair.key_type == ED25519
        assert suite.key_pair.public_key_base58 == TEST_DID_KEY_VERKEY

        suite = await self.manager._get_suite(
            proof_type=Ed25519Signature2020.signature_type,
            verification_method=TEST_DID_KEY_VM,
            proof=proof,
            did_info=did_info,
        )

        assert isinstance(suite, Ed25519Signature2020)
        assert suite.verification_method == TEST_DID_KEY_VM
        assert suite.proof == proof
        assert isinstance(suite.key_pair, WalletKeyPair)
        assert suite.key_pair.key_type == ED25519
        assert suite.key_pair.public_key_base58 == TEST_DID_KEY_VERKEY

    async def test_get_proof_purpose(self):
        purpose = self.manager._get_proof_purpose()
        assert isinstance(purpose, CredentialIssuancePurpose)

        purpose = self.manager._get_proof_purpose(
            proof_purpose=AuthenticationProofPurpose.term,
            challenge="challenge",
            domain="domain",
        )
        assert isinstance(purpose, AuthenticationProofPurpose)
        assert purpose.domain == "domain"
        assert purpose.challenge == "challenge"

        with pytest.raises(VcLdpManagerError) as context:
            self.manager._get_proof_purpose(proof_purpose=AuthenticationProofPurpose.term)
        assert "Challenge is required for" in str(context.value)

        with pytest.raises(VcLdpManagerError) as context:
            self.manager._get_proof_purpose(proof_purpose="random")
        assert "Unsupported proof purpose: random" in str(context.value)

    async def test_prepare_detail(self):
        self.options.proof_type = BbsBlsSignature2020.signature_type

        assert SECURITY_CONTEXT_BBS_URL not in self.vc.context_urls

        await self.manager.prepare_credential(self.vc, self.options)

        assert SECURITY_CONTEXT_BBS_URL in self.vc.context_urls

    async def test_prepare_detail_ed25519_2020(self):
        self.options.proof_type = Ed25519Signature2020.signature_type

        assert SECURITY_CONTEXT_ED25519_2020_URL not in self.vc.context_urls

        await self.manager.prepare_credential(self.vc, self.options)

        assert SECURITY_CONTEXT_ED25519_2020_URL in self.vc.context_urls

    async def test_issue(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            did = await wallet.create_local_did(
                method=KEY,
                key_type=ED25519,
            )
        self.vc.issuer = did.did
        self.options.proof_type = Ed25519Signature2018.signature_type
        cred = await self.manager.issue(self.vc, self.options)
        assert cred

    async def test_issue_ed25519_2020(self):
        """Ensure ed25519 2020 context added to issued cred."""
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            did = await wallet.create_local_did(
                method=KEY,
                key_type=ED25519,
            )
        self.vc.issuer = did.did
        self.options.proof_type = Ed25519Signature2020.signature_type
        cred = await self.manager.issue(self.vc, self.options)
        assert cred

    @pytest.mark.ursa_bbs_signatures
    async def test_issue_bbs(self):
        """Ensure BBS context is added to issued cred."""
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            did = await wallet.create_local_did(
                method=KEY,
                key_type=BLS12381G2,
            )
        self.vc.issuer = did.did
        self.options.proof_type = BbsBlsSignature2020.signature_type
        cred = await self.manager.issue(self.vc, self.options)
        assert cred

    async def test_get_all_suites(self):
        suites = await self.manager._get_all_proof_suites()
        assert len(suites) == 5
        types = (
            Ed25519Signature2018,
            Ed25519Signature2020,
            EcdsaSecp256r1Signature2019,
            BbsBlsSignature2020,
            BbsBlsSignatureProof2020,
        )
        for suite in suites:
            assert isinstance(suite, types)

    async def test_store(
        self,
    ):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            did = await wallet.create_local_did(
                method=KEY,
                key_type=ED25519,
            )
        self.vc.issuer = did.did
        self.options.proof_type = Ed25519Signature2018.signature_type
        cred = await self.manager.issue(self.vc, self.options)
        await self.manager.store_credential(cred, TEST_UUID)
        async with self.profile.session() as session:
            holder = session.inject(VCHolder)
            record = await holder.retrieve_credential_by_id(record_id=TEST_UUID)
        assert record



"""Key admin routes."""

import logging

from aiohttp import web
from aiohttp_apispec import docs, request_schema, response_schema
from marshmallow import fields

from ...admin.decorators.auth import tenant_authentication
from ...admin.request_context import AdminRequestContext
from ...messaging.models.openapi import OpenAPISchema
from ...wallet.error import WalletDuplicateError, WalletNotFoundError
from ..routes import WALLET_TAG_TITLE
from .manager import DEFAULT_ALG, MultikeyManager, MultikeyManagerError

LOGGER = logging.getLogger(__name__)


class CreateKeyRequestSchema(OpenAPISchema):
    """Request schema for creating a new key."""

    alg = fields.Str(
        required=False,
        metadata={
            "description": "Which key algorithm to use.",
            "example": DEFAULT_ALG,
        },
    )

    seed = fields.Str(
        required=False,
        metadata={
            "description": (
                "Optional seed to generate the key pair. "
                "Must enable insecure wallet mode."
            ),
            "example": "00000000000000000000000000000000",
        },
    )

    kid = fields.Str(
        required=False,
        metadata={
            "description": (
                "Optional kid to bind to the keypair, such as a verificationMethod."
            ),
            "example": "did:web:example.com#key-01",
        },
    )


class CreateKeyResponseSchema(OpenAPISchema):
    """Response schema from creating a new key."""

    multikey = fields.Str(
        metadata={
            "description": "The Public Key Multibase format (multikey)",
            "example": "z6MkgKA7yrw5kYSiDuQFcye4bMaJpcfHFry3Bx45pdWh3s8i",
        },
    )

    kid = fields.Str(
        metadata={
            "description": "The associated kid",
            "example": "did:web:example.com#key-01",
        },
    )


class UpdateKeyRequestSchema(OpenAPISchema):
    """Request schema for updating an existing key pair."""

    multikey = fields.Str(
        required=True,
        metadata={
            "description": "Multikey of the key pair to update",
            "example": "z6MkgKA7yrw5kYSiDuQFcye4bMaJpcfHFry3Bx45pdWh3s8i",
        },
    )

    kid = fields.Str(
        required=True,
        metadata={
            "description": (
                "New kid to bind to the key pair, such as a verificationMethod."
            ),
            "example": "did:web:example.com#key-02",
        },
    )


class UpdateKeyResponseSchema(OpenAPISchema):
    """Response schema from updating an existing key pair."""

    multikey = fields.Str(
        metadata={
            "description": "The Public Key Multibase format (multikey)",
            "example": "z6MkgKA7yrw5kYSiDuQFcye4bMaJpcfHFry3Bx45pdWh3s8i",
        },
    )

    kid = fields.Str(
        metadata={
            "description": "The associated kid",
            "example": "did:web:example.com#key-02",
        },
    )


class FetchKeyResponseSchema(OpenAPISchema):
    """Response schema from updating an existing key pair."""

    multikey = fields.Str(
        metadata={
            "description": "The Public Key Multibase format (multikey)",
            "example": "z6MkgKA7yrw5kYSiDuQFcye4bMaJpcfHFry3Bx45pdWh3s8i",
        },
    )

    kid = fields.Str(
        metadata={
            "description": "The associated kid",
            "example": "did:web:example.com#key-01",
        },
    )


@docs(tags=[WALLET_TAG_TITLE], summary="Fetch key info.")
@response_schema(FetchKeyResponseSchema, 200, description="")
@tenant_authentication
async def fetch_key(request: web.BaseRequest):
    """Request handler for fetching a key.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    multikey = request.match_info["multikey"]

    try:
        async with context.session() as session:
            key_info = await MultikeyManager(session).from_multikey(multikey=multikey)
        return web.json_response(
            key_info,
            status=200,
        )

    except (MultikeyManagerError, WalletDuplicateError, WalletNotFoundError) as err:
        return web.json_response({"message": str(err)}, status=400)


@docs(tags=[WALLET_TAG_TITLE], summary="Create a key pair")
@request_schema(CreateKeyRequestSchema())
@response_schema(CreateKeyResponseSchema, 200, description="")
@tenant_authentication
async def create_key(request: web.BaseRequest):
    """Request handler for creating a new key pair in the wallet.

    Args:
        request: aiohttp request object

    Returns:
        The Public Key Multibase format (multikey)

    """
    context: AdminRequestContext = request["context"]
    body = await request.json()

    seed = body.get("seed") or None
    kid = body.get("kid") or None
    alg = body.get("alg") or DEFAULT_ALG

    if seed and not context.settings.get("wallet.allow_insecure_seed"):
        raise MultikeyManagerError("Seed support is not enabled.")

    try:
        async with context.session() as session:
            key_info = await MultikeyManager(session).create(seed=seed, kid=kid, alg=alg)
        return web.json_response(
            key_info,
            status=201,
        )
    except (MultikeyManagerError, WalletDuplicateError, WalletNotFoundError) as err:
        return web.json_response({"message": str(err)}, status=400)


@docs(tags=[WALLET_TAG_TITLE], summary="Update a key pair's kid")
@request_schema(UpdateKeyRequestSchema())
@response_schema(UpdateKeyResponseSchema, 200, description="")
@tenant_authentication
async def update_key(request: web.BaseRequest):
    """Request handler for creating a new key pair in the wallet.

    Args:
        request: aiohttp request object

    Returns:
        The Public Key Multibase format (multikey)

    """
    context: AdminRequestContext = request["context"]
    body = await request.json()

    multikey = body.get("multikey")
    kid = body.get("kid")

    try:
        async with context.session() as session:
            key_info = await MultikeyManager(session).update(
                multikey=multikey,
                kid=kid,
            )
        return web.json_response(
            key_info,
            status=200,
        )
    except (MultikeyManagerError, WalletDuplicateError, WalletNotFoundError) as err:
        return web.json_response({"message": str(err)}, status=400)


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get("/wallet/keys/{multikey}", fetch_key, allow_head=False),
            web.post("/wallet/keys", create_key),
            web.put("/wallet/keys", update_key),
        ]
    )



"""Indy tails server interface class."""

import logging
from typing import Tuple

from ..config.injection_context import InjectionContext
from ..ledger.base import BaseLedger
from ..ledger.multiple_ledger.base_manager import BaseMultipleLedgerManager
from ..utils.http import PutError, put_file
from .base import BaseTailsServer
from .error import TailsServerNotConfiguredError

LOGGER = logging.getLogger(__name__)


class IndyTailsServer(BaseTailsServer):
    """Indy tails server interface."""

    async def upload_tails_file(
        self,
        context: InjectionContext,
        filename: str,
        tails_file_path: str,
        interval: float = 1.0,
        backoff: float = 0.25,
        max_attempts: int = 5,
    ) -> Tuple[bool, str]:
        """Upload tails file to tails server.

        Args:
            context: context with configuration settings
            filename: file name given to tails server
            tails_file_path: path to the tails file to upload
            interval: initial interval between attempts
            backoff: exponential backoff in retry interval
            max_attempts: maximum number of attempts to make

        Returns:
            Tuple[bool, str]: tuple with success status and url of uploaded
            file or error message if failed

        """
        tails_server_upload_url = context.settings.get("tails_server_upload_url")
        genesis_transactions = context.settings.get("ledger.genesis_transactions")

        if not genesis_transactions:
            ledger_manager = context.injector.inject(BaseMultipleLedgerManager)
            write_ledger = context.injector.inject(BaseLedger)
            available_write_ledgers = await ledger_manager.get_write_ledgers()
            LOGGER.debug("available write_ledgers = %s", available_write_ledgers)
            LOGGER.debug("write_ledger = %s", write_ledger)
            pool = write_ledger.pool
            LOGGER.debug("write_ledger pool = %s", pool)

            genesis_transactions = pool.genesis_txns

        if not genesis_transactions:
            raise TailsServerNotConfiguredError(
                "no genesis_transactions for writable ledger"
            )

        if not tails_server_upload_url:
            raise TailsServerNotConfiguredError(
                "tails_server_upload_url setting is not set"
            )

        upload_url = tails_server_upload_url.rstrip("/") + f"/{filename}"

        try:
            await put_file(
                upload_url,
                {"tails": tails_file_path},
                {"genesis": genesis_transactions},
                interval=interval,
                backoff=backoff,
                max_attempts=max_attempts,
            )
        except PutError as x_put:
            return (False, x_put.message)

        return True, upload_url



"""Events fired by AnonCreds interface."""

import re
from typing import NamedTuple, Optional

from ..core.event_bus import Event
from .models.revocation import RevRegDef

CRED_DEF_FINISHED_EVENT = "anoncreds::credential-definition::finished"
REV_REG_DEF_FINISHED_EVENT = "anoncreds::revocation-registry-definition::finished"
REV_LIST_FINISHED_EVENT = "anoncreds::revocation-list::finished"

CRED_DEF_FINISHED_PATTERN = re.compile(CRED_DEF_FINISHED_EVENT)
REV_REG_DEF_FINISHED_PATTERN = re.compile(REV_REG_DEF_FINISHED_EVENT)
REV_LIST_FINISHED_PATTERN = re.compile(REV_LIST_FINISHED_EVENT)


class CredDefFinishedPayload(NamedTuple):
    """Payload of cred def finished event."""

    schema_id: str
    cred_def_id: str
    issuer_id: str
    support_revocation: bool
    max_cred_num: int
    options: dict


class CredDefFinishedEvent(Event):
    """Event for cred def finished."""

    def __init__(
        self,
        payload: CredDefFinishedPayload,
    ):
        """Initialize an instance.

        Args:
            payload: CredDefFinishedPayload
        """
        self._topic = CRED_DEF_FINISHED_EVENT
        self._payload = payload

    @classmethod
    def with_payload(
        cls,
        schema_id: str,
        cred_def_id: str,
        issuer_id: str,
        support_revocation: bool,
        max_cred_num: int,
        options: Optional[dict] = None,
    ):
        """With payload."""
        payload = CredDefFinishedPayload(
            schema_id=schema_id,
            cred_def_id=cred_def_id,
            issuer_id=issuer_id,
            support_revocation=support_revocation,
            max_cred_num=max_cred_num,
            options=options,
        )
        return cls(payload)

    @property
    def payload(self) -> CredDefFinishedPayload:
        """Return payload."""
        return self._payload


class RevRegDefFinishedPayload(NamedTuple):
    """Payload of rev reg def finished event."""

    rev_reg_def_id: str
    rev_reg_def: RevRegDef
    options: dict


class RevRegDefFinishedEvent(Event):
    """Event for rev reg def finished."""

    def __init__(self, payload: RevRegDefFinishedPayload):
        """Initialize an instance.

        Args:
            payload: RevRegDefFinishedPayload
        """
        self._topic = REV_REG_DEF_FINISHED_EVENT
        self._payload = payload

    @classmethod
    def with_payload(
        cls,
        rev_reg_def_id: str,
        rev_reg_def: RevRegDef,
        options: Optional[dict] = None,
    ):
        """With payload."""
        payload = RevRegDefFinishedPayload(rev_reg_def_id, rev_reg_def, options)
        return cls(payload)

    @property
    def payload(self) -> RevRegDefFinishedPayload:
        """Return payload."""
        return self._payload


class RevListFinishedPayload(NamedTuple):
    """Payload of rev list finished event."""

    rev_reg_id: str
    revoked: list
    options: dict


class RevListFinishedEvent(Event):
    """Event for rev list finished."""

    def __init__(self, payload: RevListFinishedPayload):
        """Initialize an instance.

        Args:
            payload: RevListFinishedPayload
        """
        self._topic = REV_LIST_FINISHED_EVENT
        self._payload = payload

    @classmethod
    def with_payload(
        cls,
        rev_reg_id: str,
        revoked: list,
        options: Optional[dict] = None,
    ):
        """With payload."""
        payload = RevListFinishedPayload(rev_reg_id, revoked, options)
        return cls(payload)

    @property
    def payload(self) -> RevListFinishedPayload:
        """Return payload."""
        return self._payload



"""Abstract message queue."""

import asyncio
from abc import ABC, abstractmethod
from typing import Optional


class BaseMessageQueue(ABC):
    """Abstract message queue class."""

    @abstractmethod
    async def enqueue(self, message):
        """Enqueue a message.

        Args:
            message: The message to add to the end of the queue

        Raises:
            asyncio.CancelledError if the queue has been stopped

        """

    @abstractmethod
    async def dequeue(self, *, timeout: Optional[int] = None):
        """Dequeue a message.

        Returns:
            The dequeued message, or None if a timeout occurs

        Raises:
            asyncio.CancelledError if the queue has been stopped
            asyncio.TimeoutError if the timeout is reached

        """

    @abstractmethod
    async def join(self):
        """Wait for the queue to empty."""

    @abstractmethod
    def task_done(self):
        """Indicate that the current task is complete."""

    @abstractmethod
    def stop(self):
        """Cancel active iteration of the queue."""

    @abstractmethod
    def reset(self):
        """Empty the queue and reset the stop event."""

    def __aiter__(self):
        """Async iterator magic method."""
        return self

    async def __anext__(self):
        """Async iterator magic method."""
        try:
            message = await self.dequeue()
        except asyncio.CancelledError:
            raise StopAsyncIteration
        return message
